import logging
import os
from typing import Optional
from typing import Optional
import os
import sparknlp
from pyspark.sql.functions import monotonically_increasing_id
from sparknlp.common import AnnotatorType

from nlu.pipe.utils.audio_data_conversion_utils import AudioDataConversionUtils
from nlu.pipe.utils.data_conversion_utils import DataConversionUtils, NluDataParseException
from nlu.pipe.utils.ocr_data_conversion_utils import OcrDataConversionUtils

logger = logging.getLogger('nlu')
from nlu.pipe.pipe_logic import PipeUtils

import pandas as pd
from pydantic import BaseModel


def get_first_anno_with_output_type(pipe, out_type):
    for s in pipe.vanilla_transformer_pipe.stages:
        if hasattr(s, 'outputAnnotatorType') and s.outputAnnotatorType == out_type:
            return s
    return None


def serialize(img_path):
    with open(img_path, 'rb') as img_file:
        return img_file.read()


def deserialize(binary_image, path):
    with open(path, 'wb') as img_file:
        img_file.write(binary_image)


class PredictParams(BaseModel):
    output_level: Optional[str] = ''
    positions: Optional[bool] = False
    keep_stranger_features: Optional[bool] = True
    metadata: Optional[bool] = False
    multithread: Optional[bool] = True
    drop_irrelevant_cols: Optional[bool] = True
    return_spark_df: Optional[bool] = False
    get_embeddings: Optional[bool] = True

    @staticmethod
    def has_param_cols(df: pd.DataFrame):
        return all([c not in df.columns for c in PredictParams.__fields__.keys()])

    @staticmethod
    def maybe_from_pandas_df(df: pd.DataFrame):
        # only first row is used
        if df.shape[0] == 0:
            return None
        if PredictParams.has_param_cols(df):
            # no params in df
            return None
        param_row = df.iloc[0].to_dict()
        try:
            return PredictParams(**param_row)
        except Exception as e:
            print(f'Exception trying to parse prediction parameters for param row:'
                  f' \n{param_row} \n', e)
            return None


def get_first_anno_with_output_type(pipe, out_type):
    for s in pipe.vanilla_transformer_pipe.stages:
        if hasattr(s, 'outputAnnotatorType') and s.outputAnnotatorType == out_type:
            return s
    return None


def __predict_standard_spark(pipe, data, output_level, positions, keep_stranger_features, metadata,
                             drop_irrelevant_cols, return_spark_df, get_embeddings):
    # 1. Convert data to Spark DF
    data, stranger_features, output_datatype = DataConversionUtils.to_spark_df(data, pipe.spark, pipe.raw_text_column,
                                                                               is_span_data=pipe.has_span_classifiers,
                                                                               is_tabular_qa_data=pipe.has_table_qa_models,
                                                                               )

    # 3. Apply Spark Pipeline
    data = pipe.vanilla_transformer_pipe.transform(data)

    # 4. Convert resulting spark DF into nicer format and by default into pandas.
    if return_spark_df: return data  # Returns RAW  Spark Dataframe result of component_list prediction
    return pipe.pythonify_spark_dataframe(data,
                                          keep_stranger_features=keep_stranger_features,
                                          stranger_features=stranger_features,
                                          output_metadata=metadata,
                                          drop_irrelevant_cols=drop_irrelevant_cols,
                                          positions=positions,
                                          output_level=output_level,
                                          get_embeddings=get_embeddings
                                          )


def predict_multi_threaded_light_pipe(pipe, data, output_level, positions, keep_stranger_features, metadata,
                                      drop_irrelevant_cols, get_embeddings):
    # 1. Try light component_list predcit
    # 2. if vanilla Fails use Vanilla
    # 3.  if vanilla fails raise error
    data, stranger_features, output_datatype = DataConversionUtils.to_pandas_df(data, pipe.raw_text_column)

    # Predict -> Cast to PDF -> Join with original inputs. It does NOT yield EMBEDDINGS.
    data = data.join(pd.DataFrame(pipe.light_transformer_pipe.fullAnnotate(data.text.values.tolist())))

    return pipe.pythonify_spark_dataframe(data,
                                          keep_stranger_features=keep_stranger_features,
                                          stranger_features=stranger_features,
                                          output_metadata=metadata,
                                          drop_irrelevant_cols=drop_irrelevant_cols,
                                          positions=positions,
                                          output_level=output_level,
                                          get_embeddings=get_embeddings
                                          )


def __predict_ocr_spark(pipe, data, output_level, positions, keep_stranger_features, metadata,
                        drop_irrelevant_cols, get_embeddings):
    """
        Check if there are any OCR components in the Pipe.
        If yes, we verify data contains pointer to jsl_folder or image files.
        If yes, df = spark.read.format("binaryFile").load(imagePath)
        Run OCR pipe on df and pythonify procedure afterwards

    """
    pipe.fit()

    OcrDataConversionUtils.validate_OCR_compatible_inputs(data)
    paths = OcrDataConversionUtils.extract_iterable_paths_from_data(data)
    accepted_file_types = OcrDataConversionUtils.get_accepted_ocr_file_types(pipe)
    file_paths = OcrDataConversionUtils.glob_files_of_accepted_type(paths, accepted_file_types)
    spark = sparknlp.start()  # Fetches Spark Session that has already been licensed

    # Some annos require `image` format, some will require `binary` format. We need to figure out which one is needed possible provide both
    if pipe.requires_image_format and pipe.requires_binary_format:
        from pyspark.sql.functions import regexp_replace
        # Image & Binary formats required. We read as both and join the dfs
        img_df = spark.read.format("image").load(file_paths).withColumn("modified_origin",
                                                                        regexp_replace("image.origin", ":/{1,}", ":"))

        # Read the files in binaryFile format
        binary_df = spark.read.format("binaryFile").load(file_paths).withColumn("modified_path",
                                                                                regexp_replace("path", ":/{1,}", ":"))

        data = img_df.join(binary_df, img_df["modified_origin"] == binary_df["modified_path"]).drop('modified_path')

    elif pipe.requires_image_format:
        # only image format required
        data = spark.read.format("image").load(file_paths)
    elif pipe.requires_binary_format:
        # only binary required
        data = spark.read.format("binaryFile").load(file_paths)
    else:
        # fallback default
        data = spark.read.format("binaryFile").load(file_paths)
    data = data.withColumn('origin_index', monotonically_increasing_id().alias('origin_index'))

    data = pipe.vanilla_transformer_pipe.transform(data)
    return pipe.pythonify_spark_dataframe(data,
                                          keep_stranger_features=keep_stranger_features,
                                          output_metadata=metadata,
                                          drop_irrelevant_cols=drop_irrelevant_cols,
                                          positions=positions,
                                          output_level=output_level,
                                          get_embeddings=get_embeddings
                                          )


def __predict_audio_spark(pipe, data, output_level, positions, keep_stranger_features, metadata,
                          drop_irrelevant_cols, get_embeddings):
    """
        Check if there are any OCR components in the Pipe.
        If yes, we verify data contains pointer to jsl_folder or image files.
        If yes, df = spark.read.format("binaryFile").load(imagePath)
        Run OCR pipe on df and pythonify procedure afterwards

    """
    pipe.fit()

    try:
        import librosa
    except:
        raise ImportError('The librosa library is not installed and required for audio features! '
                          'Run pip install librosa ')
    sample_rate = 16000
    AudioDataConversionUtils.validate_paths(data)
    paths = AudioDataConversionUtils.extract_iterable_paths_from_data(data)
    accepted_file_types = AudioDataConversionUtils.get_accepted_audio_file_types(pipe)
    file_paths = AudioDataConversionUtils.glob_files_of_accepted_type(paths, accepted_file_types)
    data = AudioDataConversionUtils.data_to_spark_audio_df(data=file_paths, sample_rate=sample_rate,
                                                           spark=sparknlp.start())
    data = pipe.vanilla_transformer_pipe.transform(data).withColumn(
        'origin_index', monotonically_increasing_id().alias('origin_index'))

    return pipe.pythonify_spark_dataframe(data,
                                          keep_stranger_features=keep_stranger_features,
                                          output_metadata=metadata,
                                          drop_irrelevant_cols=drop_irrelevant_cols,
                                          positions=positions,
                                          output_level=output_level,
                                          get_embeddings=get_embeddings
                                          )


def __db_endpoint_predict__(pipe, data):
    """
    1) parse pred params from first row maybe
    2) serialize/deserialize img
    """
    params = PredictParams.maybe_from_pandas_df(data)
    if params:
        params = params.dict()
    else:
        params = {}
    files = []
    if 'file' in data.columns and 'file_type' in data.columns:
        print("DETECTED FILE COLS")
        skip_first = PredictParams.has_param_cols(data)
        for i, row in data.iterrows():
            print(f"DESERIALIZING {row.file_type} file {row.file}")
            if i == 0 and skip_first:
                continue
            file_name = f'file{i}.{row.file_type}'
            files.append(file_name)
            deserialize(row.file, file_name)
        data = files

    if params:
        return __predict__(pipe, data, **params, normal_pred_on_db=True)
    else:
        # no params detect, we call again with default params
        return __predict__(pipe, data, **PredictParams().dict(), normal_pred_on_db=True)


def __predict_standard_spark_only_embed(pipe, data, return_spark_df):
    # 1. Convert data to Spark DF
    data, stranger_features, output_datatype = DataConversionUtils.to_spark_df(data, pipe.spark, pipe.raw_text_column,
                                                                               is_span_data=pipe.has_span_classifiers,
                                                                               is_tabular_qa_data=pipe.has_table_qa_models,
                                                                               )

    # 2. Apply Spark Pipeline
    data = pipe.vanilla_transformer_pipe.transform(data)

    # 3. Validate data
    sent_embedder = get_first_anno_with_output_type(pipe, AnnotatorType.SENTENCE_EMBEDDINGS)
    if not sent_embedder or not hasattr(sent_embedder, 'getOutputCol', ):
        raise Exception('No Sentence Embedder found in pipeline')

    # 4. return embeds
    emb_col = sent_embedder.getOutputCol()
    if return_spark_df:
        return data.select(f'{emb_col}.embeddings')

    # Note, this is only document output level. If pipe has sentence detector, we will only keep first embed of every document.
    return [r.embeddings[0] for r in data.select(f'{emb_col}.embeddings').collect()]

def try_update_session():
    try:
        import sparknlp
        spark = sparknlp.start()
        spark._jvm.com.johnsnowlabs.license.LicenseValidator.meterServingUsage(
            spark._jvm.scala.Option.apply(None)
        )
    except Exception as e:
        print(f"Error updating session: {e}")

def __predict__(pipe, data, output_level, positions, keep_stranger_features, metadata, multithread,
                drop_irrelevant_cols, return_spark_df, get_embeddings, embed_only=False,normal_pred_on_db=False):
    '''
    Annotates a Pandas Dataframe/Pandas Series/Numpy Array/Spark DataFrame/Python List strings /Python String
    :param data: Data to predict on
    :param output_level: output level, either document/sentence/chunk/token
    :param positions: whether to output indexes that map predictions back to position in origin string
    :param keep_stranger_features: whether to keep columns in the dataframe that are not generated by pandas. I.e. when you s a dataframe with 10 columns and only one of them is named text, the returned dataframe will only contain the text column when set to false
    :param metadata: whether to keep additional metadata in final df or not like confidences of every possible class for preidctions.
    :param multithread: Whether to use multithreading based light pipeline. In some cases, this may cause errors.
    :param drop_irellevant_cols: Whether to drop cols of different output levels, i.e. when predicting token level and dro_irrelevant_cols = True then chunk, sentence and Doc will be dropped
    :param return_spark_df: Prediction results will be returned right after transforming with the Spark NLP pipeline
    :return:
    '''
    if embed_only:
        pipe.fit()
        return __predict_standard_spark_only_embed(pipe, data, return_spark_df)

    if 'DB_ENDPOINT_ENV' in os.environ and not normal_pred_on_db:
        print('DEBUG DB ENV: PREDICIOTN ON DB ENDPOINT')
        print(data)
        print(type(data))
        try_update_session()
        return __db_endpoint_predict__(pipe,data)

    if output_level == '' and not pipe.has_table_qa_models:
        # Default sentence level for all components
        if pipe.has_nlp_components and not PipeUtils.contains_t5_or_gpt(
                pipe) and not pipe.has_span_classifiers:
            pipe.component_output_level = 'sentence'
            pipe.components = PipeUtils.configure_component_output_levels(pipe, 'sentence')
    else:
        if pipe.has_nlp_components and output_level in ['document', 'sentence']:
            # Pipe must be re-configured for document/sentence level
            pipe.component_output_level = output_level
            pipe.components = PipeUtils.configure_component_output_levels(pipe, output_level)

        elif pipe.has_nlp_components and output_level in ['token']:
            # Add tokenizer if not iadel = 'sentence'
            pipe.components = PipeUtils.configure_component_output_levels(pipe, 'sentence')
            pipe = PipeUtils.add_tokenizer_to_pipe_if_missing(pipe)

    if get_embeddings is None:
        # Grab embeds if nlu ref is of type embed
        get_embeddings = True if 'embed' in pipe.nlu_ref else False

    if not pipe.is_fitted:
        if pipe.has_trainable_components:
            pipe.fit(data)
        else:
            pipe.fit()

        if not pipe.is_light_pipe_incompatible:
            pipe.__configure_light_pipe_usage__(DataConversionUtils.size_of(data), multithread)

    if pipe.contains_ocr_components and pipe.contains_audio_components:
        """ Idea:
        Expect Array of Paths 
        For every path classify file ending and use it to correctly handle Img or Audio stuff 
        """
        raise Exception('Cannot mix Audio and OCR components in a Pipe?')

    if pipe.contains_audio_components:
        return __predict_audio_spark(pipe, data, output_level, positions, keep_stranger_features,
                                     metadata, drop_irrelevant_cols, get_embeddings=get_embeddings)

    if pipe.contains_ocr_components:
        # Ocr processing
        try:
            return __predict_ocr_spark(pipe, data, output_level, positions, keep_stranger_features,
                                       metadata, drop_irrelevant_cols, get_embeddings=get_embeddings)
        except Exception as err:
            logger.warning(f"Predictions Failed={err}")
            pipe.print_exception_err(err)
            raise Exception("Failure to process data with NLU OCR pipe")
    if return_spark_df:
        try:
            return __predict_standard_spark(pipe, data, output_level, positions, keep_stranger_features, metadata,
                                            drop_irrelevant_cols, return_spark_df, get_embeddings)
        except Exception as err:
            logger.warning(f"Predictions Failed={err}")
            pipe.print_exception_err(err)
            raise Exception("Failure to process data with NLU")
    elif not get_embeddings and multithread or pipe.prefer_light:
        # In Some scenarios we prefer light, because Bugs in ChunkMapper...
        # Try Multithreaded with Fallback vanilla as option. No Embeddings in this mode
        try:
            return predict_multi_threaded_light_pipe(pipe, data, output_level, positions, keep_stranger_features,
                                                     metadata, drop_irrelevant_cols, get_embeddings=get_embeddings)


        except Exception as err:
            logger.warning(
                f"Multithreaded mode with Light pipeline failed. trying to predict again with non multithreaded mode, "
                f"err={err}")
            try:
                return __predict_standard_spark(pipe, data, output_level, positions, keep_stranger_features, metadata,
                                                drop_irrelevant_cols, return_spark_df, get_embeddings)
            except Exception as err:
                logger.warning(f"Predictions Failed={err}")
                pipe.print_exception_err(err)
                raise Exception("Failure to process data with NLU")
    else:
        # Standard predict with no fallback
        try:
            return __predict_standard_spark(pipe, data, output_level, positions, keep_stranger_features, metadata,
                                            drop_irrelevant_cols, return_spark_df, get_embeddings)
        except NluDataParseException as err:
            logger.warning(f"Predictions Failed={err}")
            raise err
        except Exception as err:
            logger.warning(f"Predictions Failed={err}")
            pipe.print_exception_err(err)
            raise Exception("Failure to process data with NLU")

def debug_print_pipe_cols(pipe):
    for c in pipe.components:
        print(f'{c.spark_input_column_names}->{c.name}->{c.spark_output_column_names}')



class Doc2Text:
    @staticmethod
    def get_default_model():
        from sparkocr.transformers import DocToText
        return DocToText() \
            .setInputCol("content") \
            .setOutputCol("text")



from pyspark.sql.types import *
import glob
import logging
import os
from typing import List
import pyspark
import numpy as np
import pandas as pd

logger = logging.getLogger('nlu')


class AudioDataConversionUtils:
    """Validate Audio Data Files and Create Spark DataFrames from them"""

    @staticmethod
    def validate_paths(data):
        """Validate for input data that it contains a path pointing to file or folder of audio fila readable with librosa"""
        if isinstance(data, List):
            return AudioDataConversionUtils.check_iterable_paths_are_valid(data)
        if isinstance(data, str):
            return AudioDataConversionUtils.check_iterable_paths_are_valid([data])
        if isinstance(data, pd.DataFrame):
            return 'path' in data.columns
        if isinstance(data, pd.Series):
            return 'path' in data.name
        if isinstance(data, pyspark.sql.dataframe.DataFrame):
            return 'path' in data.columns
        if isinstance(data, np.ndarray):
            return AudioDataConversionUtils.check_iterable_paths_are_valid(data)

    @staticmethod
    def check_iterable_paths_are_valid(iterable_paths):
        """Validate for iterable data input if all elements point to file or jsl_folder"""
        paths_validness = []
        for p in iterable_paths:
            if os.path.isdir(p) or os.path.isfile(p):
                paths_validness.append(True)
            else:
                print(f'Warning : Invalid path for jsl_folder or file in input. Could validate path.\n'
                      f'NLU will try and ignore this issue but you might run into errors.\n'
                      f'Please make sure all paths are valid\n')
                print(f'For path = {p}')
                paths_validness.append(False)
        return all(paths_validness)

    @staticmethod
    def check_all_paths_point_to_accepted_file_type(paths, file_types):
        """Validate that all paths point to a file type defined by file_types"""
        pass

    @staticmethod
    def data_to_spark_audio_df(data, sample_rate, spark):
        import librosa
        import typing
        if isinstance(data, str):
            data, _ = librosa.load(data, sr=sample_rate)
            # let's convert them to floats
            df = pd.DataFrame({
                # data is an List[List[Float]]
                "raw_audio": data.toList(),
                "sampling_rate": [sample_rate]
            })
        elif isinstance(data, typing.Iterable):
            data_ = []
            for d in data:
                data, _ = librosa.load(d, sr=sample_rate)
                data_.append(data.tolist())
            df = pd.DataFrame({
                # data is an List[List[Float]]
                "raw_audio": data_,
                "sampling_rate": [sample_rate] * len(data_)
            })
        schema = StructType([StructField("raw_audio", ArrayType(FloatType())),
                             StructField("sampling_rate", LongType())])
        data = spark.createDataFrame(df, schema)
        return data

    @staticmethod
    def glob_files_of_accepted_type(paths, file_types):
        """Get all paths which point to correct file types from iterable paths which can contain file and jsl_folder paths
        1. paths point to a file which is suffixed with one of the accepted file_types, i.e. path/to/file.type
        2. path points to a jsl_folder, in this case jsl_folder is recursively searched for valid files and accepted paths will be in return result
        """
        accepted_file_paths = []
        for p in paths:
            for t in file_types:
                t = t.lower()
                if os.path.isfile(p):
                    if p.lower().split('.')[-1] == t:
                        accepted_file_paths.append(p)
                elif os.path.isdir(p):
                    accepted_file_paths += glob.glob(p + f'/**/*.{t}', recursive=True)
                else:
                    print(f"Invalid path = {p} pointing neither to file or jsl_folder on this machine")
        return accepted_file_paths

    @staticmethod
    def extract_iterable_paths_from_data(data):
        """Extract an iterable object containing paths from input data"""
        if isinstance(data, List):
            return data
        if isinstance(data, str):
            return [data]
        if isinstance(data, pd.DataFrame):
            return list(data['path'].values)
        if isinstance(data, pd.Series):
            return list(data.values)
        if isinstance(data, pyspark.sql.dataframe.DataFrame):
            return [p['path'] for p in data.select('path').collect()]
        if isinstance(data, np.ndarray):
            return list(data)

    @staticmethod
    def get_accepted_audio_file_types(pipe):
        """Get all file types/suffixes that can be processed by the pipeline"""
        accepted_files = []
        for c in pipe.components:
            if c.applicable_file_types:
                accepted_files += c.applicable_file_types
        return list(set(accepted_files))



from sparknlp.annotator import *


class DistilBertZeroShotClassifier:
    @staticmethod
    def get_default_model():
        return DistilBertForZeroShotClassification.pretrained() \
            .setInputCols(["token", "sentence"]) \
            .setOutputCol("category") \
            .setCaseSensitive(True)

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return DistilBertForZeroShotClassification.pretrained(name, language, bucket) \
            .setInputCols(["token", "sentence"]) \
            .setOutputCol("category") \
            .setCaseSensitive(True)



from sparknlp.annotator import *

class LanguageDetector:
    @staticmethod
    def get_default_model():
        return LanguageDetectorDL.pretrained()\
        .setInputCols("document") \
        .setOutputCol("language")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return LanguageDetectorDL.pretrained(name,language,bucket) \
            .setInputCols("document") \
            .setOutputCol("language")



class Image2Hocr:
    @staticmethod
    def get_default_model():
        from sparkocr.transformers import ImageToHocr
        return ImageToHocr() \
            .setInputCol("image") \
            .setOutputCol("hocr")



class ImageTableCellDetector:
    @staticmethod
    def get_default_model():
        from sparkocr.transformers import ImageTableCellDetector
        return ImageTableCellDetector() \
            .setInputCol("image_region") \
            .setAlgoType("morphops") \
            .setOutputCol("ocr_table_cells")



from sparknlp.annotator import XlmRoBertaForSequenceClassification


class SeqXlmRobertaClassifier:
    @staticmethod
    def get_default_model():
        return XlmRoBertaForSequenceClassification.pretrained() \
            .setInputCols(["token", "sentence"]) \
            .setOutputCol("category")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return XlmRoBertaForSequenceClassification.pretrained(name, language, bucket) \
            .setInputCols(["token", "sentence"]) \
            .setOutputCol("category")





class ChunkResolver:
    @staticmethod
    def get_pretrained_model(name, language, bucket='clinical/models'):
        from sparknlp_jsl.annotator import ChunkEntityResolverModel,ChunkEntityResolverApproach
        return ChunkEntityResolverModel.pretrained(name, language,bucket) \
            .setInputCols(["token","chunk_embeddings"]) \
            .setNeighbours(3) \
            .setOutputCol("chunk_resolution")

    @staticmethod
    def get_default_trainable_model():
        from sparknlp_jsl.annotator import ChunkEntityResolverModel,ChunkEntityResolverApproach
        return ChunkEntityResolverApproach() \
        .setInputCols("token","chunk_embeddings") \
        .setOutputCol("chunk_resolution") \
                .setLabelCol('y')\
        .setNormalizedCol("_y") \
            .setNeighbours(1000) \
            .setAlternatives(25) \
            .setEnableWmd(True).setEnableTfidf(True).setEnableJaccard(True) \
            .setEnableSorensenDice(True).setEnableJaroWinkler(True).setEnableLevenshtein(True) \
            .setDistanceWeights([1, 2, 2, 1, 1, 1]) \
            .setAllDistancesMetadata(True) \
            .setPoolingStrategy("MAX") \
            .setThreshold(1e32)
        # .setDistanceFunction("COSINE") \



from typing import Dict, Callable, Optional, List, Any, Union

from sparknlp.common import AnnotatorApproach, AnnotatorModel
from sparknlp.internal import AnnotatorTransformer

from nlu.universe.atoms import NlpLevel, LicenseType, JslAnnoId, JslAnnoPyClass, LanguageIso, \
    JslFeature
from nlu.universe.feature_node_universes import NlpFeatureNode
from nlu.universe.universes import ComponentBackends


def debug_print_pipe_cols(pipe):
    for c in pipe.components:
        print(f'{c.spark_input_column_names}->{c.name}->{c.spark_output_column_names}')


class NluComponent:
    """Contains various metadata about the loaded component_to_resolve"""

    def __init__(self,
                 name: str,  # Name for this anno
                 type: str,  # this tells us which kind of component_to_resolve this is
                 # Extractor method applicable to Pandas DF for getting pretty outputs
                 pdf_extractor_methods: Dict[str, Callable[[], any]],
                 pdf_col_name_substitutor: Callable[[], any],
                 # substitution method for renaming final cols to somthing redable
                 # sdf_extractor_methods : Dict[str,Callable[[],any]] # Extractor method applicable to Spark  DF for getting pretty outputs # TODO
                 # sdf_col_name_substitutor : Optional[Callable[[],any]] # substitution method for renaming final cols to somthing redable # TODO
                 output_level: NlpLevel,
                 # Output level of the component_to_resolve for data transformation logic or call it putput mapping??
                 node: NlpFeatureNode,  # Graph node
                 description: str,  # general annotator/model_anno_obj/component_to_resolve/pipeline info
                 provider: ComponentBackends,
                 # Who provides the implementation of this annotator, Spark-NLP for base. Would be
                 license: LicenseType,  # open source or private
                 computation_context: str,
                 # Will this component_to_resolve do its computation in Spark land (like all of Spark NLP annotators do) or does it require some other computation engine or library like Tensorflow, Numpy, HuggingFace, etc..
                 output_context: str,  # Will this components final result
                 jsl_anno_class_id: JslAnnoId,  # JSL Annotator Class this belongs to
                 jsl_anno_py_class: JslAnnoPyClass,  # JSL Annotator Class this belongs to
                 get_default_model: Optional[Callable[[], AnnotatorTransformer]] = None,
                 # Returns Concrete JSL Annotator object.
                 # Returns Concrete JSL Annotator object. May by None lang,name, bucket
                 get_pretrained_model: Optional[Callable[[str, str, str], AnnotatorTransformer]] = None,
                 # Returns Concrete JSL Annotator object. May by None
                 get_trainable_model: Optional[Callable[[], AnnotatorTransformer]] = None,
                 trainable: bool = False,
                 language: [LanguageIso] = None,
                 # constructor_args: ComponentConstructorArgs = None  # Args used to originally create this component_to_resolve
                 nlu_ref: str = None,
                 nlp_ref: str = None,
                 in_types: List[JslFeature] = None,
                 out_types: List[JslFeature] = None,
                 in_types_default: List[JslFeature] = None,
                 out_types_default: List[JslFeature] = None,
                 spark_input_column_names: List[str] = None,
                 spark_output_column_names: List[str] = None,
                 # Any anno class. Needs to be Any, so we can cover unimported HC models
                 model: Union[AnnotatorApproach, AnnotatorModel] = None,
                 storage_ref: Optional[str] = None,
                 storage_ref_nlu_ref_resolution: Optional[str] = None,  # nlu_ref corresponding to storage_ref
                 loaded_from_pretrained_pipe: bool = False,
                 # If this component_to_resolve was derived from a pre-built SparkNLP pipeline or from NLU
                 has_storage_ref: bool = False,
                 is_storage_ref_consumer: bool = False,
                 # # Whether this anno takes in some features that are storage ref based
                 is_storage_ref_producer: bool = False,
                 # Whether this anno generates some features that are storage ref based
                 # Reference to trainable version of this anno, if this is a non-trainable anno otherwise None
                 trainable_mirror_anno: Optional[JslAnnoId] = None,
                 # Reference to trained version of this anno, if this is a trainable anno otherwise None
                 trained_mirror_anno: Optional[JslAnnoId] = None,
                 applicable_file_types: List[str] = None,  # Used for OCR annotators to deduct applicable file types
                 is_trained: bool = True,  # Set to true for trainable annotators
                 requires_binary_format: bool = False,  # Set to true for OCR annotators that require binary image format
                 requires_image_format: bool = False,  # Set to true for OCR annotators that require image format
                 is_visual_annotator: bool = False,  # Set to true for OCR annotators that require image format
                 is_light_pipe_incompatible: bool = False,  # Set to true for OCR annotators that require image format
                 prefer_light_pipe: bool = False,  # Set True for annos that should run in light pipe
                 ):
        self.name = name
        self.type = type
        self.pdf_extractor_methods = pdf_extractor_methods
        self.pdf_col_name_substitutor = pdf_col_name_substitutor
        self.output_level = output_level
        self.node = node
        self.description = description
        self.provider = provider
        self.license = license
        self.computation_context = computation_context
        self.output_context = output_context
        self.jsl_anno_class_id = jsl_anno_class_id
        self.jsl_anno_py_class = jsl_anno_py_class
        self.get_default_model = get_default_model
        self.get_pretrained_model = get_pretrained_model
        self.get_trainable_model = get_trainable_model
        self.trainable = trainable
        self.language = language
        self.nlu_ref = nlu_ref
        self.nlp_ref = nlp_ref
        self.in_types = in_types
        self.out_types = out_types
        self.in_types_default = in_types_default
        self.out_types_default = out_types_default
        self.spark_input_column_names = spark_input_column_names
        self.spark_output_column_names = spark_output_column_names
        self.model = model
        self.storage_ref = storage_ref
        self.storage_ref_nlu_ref_resolution = storage_ref_nlu_ref_resolution
        self.loaded_from_pretrained_pipe = loaded_from_pretrained_pipe
        self.has_storage_ref = has_storage_ref
        self.is_storage_ref_consumer = is_storage_ref_consumer
        self.is_storage_ref_producer = is_storage_ref_producer
        self.trainable_mirror_anno = trainable_mirror_anno
        self.trained_mirror_anno = trained_mirror_anno
        self.applicable_file_types = applicable_file_types
        self.is_trained = is_trained
        self.requires_binary_format = requires_binary_format
        self.requires_image_format = requires_image_format
        self.is_visual_annotator = is_visual_annotator
        self.is_light_pipe_incompatible = is_light_pipe_incompatible
        self.prefer_light_pipe = prefer_light_pipe

    def set_metadata(self, jsl_anno_object: Union[AnnotatorApproach, AnnotatorModel],
                     nlu_ref: str,
                     nlp_ref: str,
                     language: LanguageIso,
                     loaded_from_pretrained_pipe: bool,
                     license_type: Optional[LicenseType] = None,
                     storage_ref: Optional[str] = None):
        """Write metadata to nlu component_to_resolve after constructing it """
        self.model = jsl_anno_object
        # converted pipes from users may have no refs attached, so we use uuid for now (todo generate nicer names based on jsl_anno_id and if its unique in pipe or not and using storage_ref)
        self.nlu_ref = nlu_ref if nlu_ref else str(jsl_anno_object)
        self.nlp_ref = nlp_ref if nlp_ref else str(jsl_anno_object)
        self.language = language if language else 'en'
        self.loaded_from_pretrained_pipe = loaded_from_pretrained_pipe
        self.in_types = self.node.ins.copy()
        self.out_types = self.node.outs.copy()
        self.in_types_default = self.node.ins.copy()
        self.out_types_default = self.node.outs.copy()
        self.spark_input_column_names = self.in_types.copy()
        self.spark_output_column_names = self.out_types.copy()
        if storage_ref:
            self.storage_ref = storage_ref
        if license_type:
            self.license = license_type
        if nlp_ref == 'glove_840B_300' or nlp_ref == 'glove_6B_300':
            self.lang = 'xx'
        if hasattr(self.model, 'setIncludeConfidence'):
            self.model.setIncludeConfidence(True)
        # if self.has_storage_ref and 'converter' in self.name:
        from nlu.universe.feature_node_ids import NLP_NODE_IDS
        if self.name in [NLP_NODE_IDS.SENTENCE_EMBEDDINGS_CONVERTER, NLP_NODE_IDS.CHUNK_EMBEDDINGS_CONVERTER]:
            # Embedding converters initially have
            self.storage_ref = ''
        if self.trainable:
            self.is_trained = False

        return self

    def set_input(self, input_cols: Union[str, List[str]]):
        if hasattr(self.model, 'setInputCol'):
            self.model.setInputCol(input_cols)
        elif hasattr(self.model, 'setInputCols'):
            self.model.setInputCols(input_cols)
        else:
            raise Exception(f'Cannot set input cols on {self.model}')
        if isinstance(input_cols, str):
            self.spark_input_column_names = [input_cols]
        elif isinstance(input_cols, list):
            self.spark_input_column_names = input_cols
        return self

    def set_output(self, output_cols: Union[str, List[str]]):
        if hasattr(self.model, 'setOutputCol'):
            self.model.setOutputCol(output_cols)
        elif hasattr(self.model, 'setOutputCols'):
            self.model.setOutputCols(output_cols)
        else:
            raise Exception(f'Cannot set output cols on {self.model}')

        if isinstance(output_cols, str):
            self.spark_output_column_names = [output_cols]
        elif isinstance(output_cols, list):
            self.spark_output_column_names = output_cols
        return self

    def __str__(self):
        return f'Component(ID={self.name}, NLU_REF={self.nlu_ref} NLP_REF={self.nlp_ref})'

    def __hash__(self):
        return hash((self.name, self.nlu_ref, self.nlp_ref, self.jsl_anno_class_id, self.language))



import glob
import json
import logging
import os
import sys
from dataclasses import dataclass
import nlu

COMPONENT_INFO_FILE_NAME = 'component_infos.json'
logger = logging.getLogger('nlu')


class AllComponentsInfo:
    def __init__(self):
        ''' Initialize every NLU component_to_resolve info object and provide access to them'''

        self.all_components = {}
        self.classifiers = {}
        self.embeddings = {}
        self.normalizers = {}
        self.pretrained_pipelines = {}
        self.selectors = {}
        self.spell_checkers = {}
        self.stemmers = {}
        self.tokenizers = {}
        self.utils = {}
        self.all_multi_lang_base_ner_languages = ['en', 'fr', 'de', 'it', 'pl', 'pt', 'ru', 'es']
        self.all_multi_lang_xtreme_ner_languages = ['af', 'ar', 'bg', 'bn', 'de', 'el', 'en', 'es', 'et', 'eu', 'fa',
                                                    'fi', 'fr', 'he', 'hi', 'hu', 'id', 'it', 'ja', 'jv', 'ka', 'kk',
                                                    'ko', 'ml', 'mr', 'ms', 'my', 'nl', 'pt', 'ru', 'sw', 'ta', 'te',
                                                    'th', 'tl', 'tr', 'ur', 'vi', 'yo', 'zh']
        self.all_right_to_left_langs_with_pretrained_tokenizer = ['zh', 'ko', 'ja']
        self.all_pretrained_pipe_languages = ['en', 'nl', 'fr', 'de', 'it', 'no', 'pl', 'pt', 'ru', 'es', 'xx', ]
        self.all_pretrained_model_languages = ['vi', 'mt', 'ta', 'af', 'cy', 'et', 'bh', 'am', 'da', 'fr', 'de', 'it',
                                               'nb', 'no', 'nn', 'pl', 'pt', 'ru', 'es', 'af', 'ar', 'hy', 'eu', 'bn',
                                               'br', 'bg', 'ca', 'cs', 'eo', 'fi', 'gl', 'el', 'ha', 'he', 'hi', 'hu',
                                               'id', 'ga', 'ja', 'la', 'lv', 'mr', 'fa', 'ro', 'sk', 'sl', 'so', 'st',
                                               'sw', 'sv', 'th', 'tr', 'uk', 'yo', 'zu', 'zh', 'xx', 'ur', 'ko',
                                               'yi', 'uk', 'te', 'ta', 'sd', 'pa', 'ne', 'ml', 'mr', 'kn', 'id', 'gu',
                                               'bs',
                                               'ig', 'lg', 'lou', 'pcm', 'wo', 'rw', 'is',
                                               ] + self.all_multi_lang_xtreme_ner_languages
        self.all_languages = set(self.all_pretrained_pipe_languages).union(set(self.all_pretrained_model_languages))
        self.all_classifier_classes = []
        # this maps a requested token to a class
        self.all_nlu_actions = ['tokenize', 'pos', 'ner', 'embed', 'classify', 'sentiment', 'emotion', 'spell',
                                'dependency', 'dep', 'dep.untyped', 'match', 'sentence_detector', 'spell', 'stopwords'
                                                                                                           'labled_dependency',
                                'lemma', 'norm', 'select', 'pretrained_pipe', 'util', 'embed_sentence', 'embed_chunk',
                                'ngram']

        # all_component_paths_regex = nlu.nlu_package_location + 'components/*/*/'
        # all_component_paths = glob.glob(all_component_paths_regex)

        # for path in all_component_paths:
        #     if '__py' in path: continue
        #     # logger.info('Loading info dict @ path'+ path)
        #     component = ComponentInfo.from_directory(path)
        #     self.all_components[component.name] = component
        #     if component.type == 'classifier': self.classifiers[component.name] = component
        #     if component.type == 'embedding': self.embeddings[component.name] = component
        #     if component.type == 'normalizer': self.normalizers[component.name] = component
        #     if component.type == 'pretrained_pipeline': self.pretrained_pipelines[component.name] = component
        #     if component.type == 'selector': self.selectors[component.name] = component
        #     if component.type == 'spell_checker': self.spell_checkers[component.name] = component
        #     if component.type == 'stemmer': self.stemmers[component.name] = component
        #     if component.type == 'tokenizer': self.tokenizers[component.name] = component
        #     if component.type == 'util': self.utils[component.name] = component

    def list_all_components(self):
        print("--------------Avaiable Components in NLU :--------------")
        for name in self.all_components.keys(): print(name)

    def get_component_info_by_name(self, name):
        return self.all_components[name]

    def list_all_components_of_type(self, component_type='embeddings'):
        pass

    @staticmethod
    def list_all_components_of_language(component_lang='ger'):
        pass

    @staticmethod
    def list_all_components_of_languageand_type(component_lang='ger', component_type='embeddings'):
        pass

    @staticmethod
    def get_default_component_of_type():
        pass

    @staticmethod
    def list_avaiable_output_types():
        pass

    @staticmethod
    def get_all_component_info_obj():
        pass


@dataclass
class ComponentInfo:
    name: str
    description: str  # general annotator/model_anno_obj/component_to_resolve/pipeline info
    outputs: list  # this is which columns/output types this component_to_resolve is providing
    inputs: list  # this tells us which columns/input types the component_to_resolve is depending on
    type: str  # this tells us which kind of component_to_resolve this is
    output_level: str  # document, sentence, token, chunk, input_dependent or model_dependent
    spark_input_column_names: list  # default expected name for input columns when forking with spark nlp annotators on spark DFs
    spark_output_column_names: list  # default expected name for output columns when forking with spark nlp annotators on spark DFs

    provider: str  # Who provides the implementation of this annotator, Spark-NLP for base. Would be
    license: str  # open source or private
    computation_context: str  # Will this component_to_resolve do its computation in Spark land (like all of Spark NLP annotators do) or does it require some other computation engine or library like Tensorflow, Numpy, HuggingFace, etc..
    output_context: str  # Will this components final result
    trainable: bool

    @classmethod
    def from_directory(cls, component_info_dir):
        """Create ComponentInfo  class from the component_infos.json which is provided for every component_to_resolve
        @param component_info_dir:
            dataset_info_dir: `str` The directory containing the metadata file. This
            should be the root directory of a specific dataset version.
        """
        if not component_info_dir:
            raise ValueError("Calling DatasetInfo.from_directory() with undefined dataset_info_dir.")
        component_info_dir = component_info_dir.replace('//', '/')
        with open(os.path.join(component_info_dir, COMPONENT_INFO_FILE_NAME), "r", encoding="utf8") as f:
            dataset_info_dict = json.load(f)

        try:
            return cls(**dataset_info_dict)  # dataset_info_dict
        except:
            print(" Exception Occured! For Path", component_info_dir,
                  " Json file most likely has missing  features. Todo nicer output error info", sys.exc_info()[0])
            raise



from typing import Optional
import streamlit as st
from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
from nlu.pipe.viz.streamlit_viz.gen_streamlit_code import get_code_for_viz
from nlu.pipe.viz.streamlit_viz.styles import _set_block_container_style
from nlu.pipe.viz.streamlit_viz.streamlit_viz_tracker import StreamlitVizTracker


class DepTreeStreamlitBlock():

    @staticmethod
    def visualize_dep_tree(
            pipe,  # nlu component_list
            text: str = 'Billy likes to swim',
            title: Optional[str] = "Dependency Parse & Part-of-speech tags",
            sub_title: Optional[
                str] = 'POS tags define a `grammatical label` for `each token` and the `Dependency Tree` classifies `Relations between the tokens` ',
            set_wide_layout_CSS: bool = True,
            generate_code_sample: bool = False,
            key="NLU_streamlit_dep_tree",
            show_infos: bool = True,
            show_logo: bool = True,
            show_text_input: bool = True,
    ):
        StreamlitVizTracker.footer_displayed = False
        if show_logo: StreamlitVizTracker.show_logo()
        if set_wide_layout_CSS: _set_block_container_style()
        if title: st.header(title)
        if show_text_input:
            text = st.text_area("Enter text you want to visualize dependency tree for ", text, key=key)
        if sub_title: st.subheader(sub_title)
        if generate_code_sample: st.code(get_code_for_viz('TREE', StreamlitUtilsOS.extract_name(pipe), text))
        if isinstance(text, str) and '\n' in text: text = text.split('\n')
        pipe.viz(text, write_to_streamlit=True, viz_type='dep', streamlit_key=key)
        if show_infos:
            # VizUtilsStreamlitOS.display_infos()
            StreamlitVizTracker.display_model_info(pipe.nlu_ref, pipes=[pipe])
            StreamlitVizTracker.display_footer()



from sparknlp.annotator import *

class SymmetricSpellChecker:
    @staticmethod
    def get_default_model():
        return SymmetricDeleteModel.pretrained() \
            .setInputCols(["token"]) \
            .setOutputCol("spell")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return SymmetricDeleteModel.pretrained(name, language) \
            .setInputCols(["token"]) \
            .setOutputCol("spell")

    @staticmethod
    def get_default_trainable_model():
        return SymmetricDeleteApproach() \
            .setInputCols(["token"]) \
            .setOutputCol("spell")



class GenericClassifier:
    @staticmethod
    def get_default_model():
        from sparknlp_jsl.annotator import  GenericClassifierModel
        return GenericClassifierModel.pretrained() \
            .setInputCols("feature_vector") \
            .setOutputCol("generic_classification") \


    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        from sparknlp_jsl.annotator import  GenericClassifierModel
        return GenericClassifierModel.pretrained(name,language,bucket) \
            .setInputCols("feature_vector") \
            .setOutputCol("generic_classification") \


    @staticmethod
    def get_default_trainable_model():
        from sparknlp_jsl.annotator import  GenericClassifierApproach
        return GenericClassifierApproach() \
            .setInputCols("feature_vector") \
            .setOutputCol("generic_classification") \
            .setLabelColumn("y") \
            .setEpochsNumber(2)



"""
Collection of universes shared across all libraries (NLP/HC/OCR), which are collections of atoms
"""
from nlu.universe.atoms import JslFeature


### ____ Annotator Feature Representations ____

class NLP_FEATURES(JslFeature):
    """
    NLP Features
    """
    #  STems from an nlp annotator in the NLP lib, i.e. Fnisher or so. Generates NO JSL-Annotation Schema for result df. Just 1 str per orw
    UNKOWN = JslFeature("unkown")

    DOCUMENT = JslFeature("document")
    DOCUMENT_FROM_CHUNK = JslFeature("document_from_chunk")
    DOCUMENT_DE_IDENTIFIED = JslFeature("document_de_identified")
    DOCUMENT_RE_IDENTIFIED = JslFeature("document_re_identified")
    DOCUMENT_NORMALIZED = JslFeature("document_normalized")
    DOCUMENT_TRANSLATED = JslFeature("document_translated")


    RAW_AUDIO = JslFeature("raw_audio")
    AUDIO = JslFeature("audio")
    RECOGNIZED_SPEECH_TEXT = JslFeature("recognized_speech_text")
    ASSEMBLED_TABULAR_DATA = JslFeature("assembled_tabular_data")

    TABULAR_ANSWER = JslFeature("tabular_answer")
    RAW_QUESTION = JslFeature("question")
    # RAW_TABULAR_DATA = JslFeature("tabular_data") # Json or csv
    RAW_QUESTION_CONTEXT = JslFeature("context")
    DOCUMENT_QUESTION = JslFeature("document_question")
    DOCUMENT_QUESTION_CONTEXT = JslFeature("document_question_context")
    CLASSIFIED_SPAN = JslFeature("classified_span")




    # GPT, T5, X2IMG (PDF2IMG, IMG2IMG, etc..)
    DOCUMENT_GENERATED = JslFeature("document_generated")

    SENTENCE = JslFeature("sentence")
    QUESTION_SENTENCES = JslFeature("question_sentences")


    TOKEN = JslFeature("token")
    COREF_TOKEN = JslFeature("coref_token")

    TOKEN_CHUNKED = JslFeature("token_chunked")
    TOKEN_SPELL_CHECKED = JslFeature("token_chunked")
    TOKEN_LEMATIZED = JslFeature("token_lemmatized")
    TOKEN_STEMMED = JslFeature("token_stemmed")
    TOKEN_NORMALIZED = JslFeature("token_stemmed")
    TOKEN_STOP_WORD_REMOVED = JslFeature("token_stemmed")


    WORDPIECE = JslFeature("wordpiece")
    ANY = JslFeature("any")
    ANY_FINISHED = JslFeature("any_finished")
    ANY_EMBEDDINGS = JslFeature("any_embeddings")
    FINISHED_EMBEDDINGS = JslFeature("word_embeddings")
    WORD_EMBEDDINGS = JslFeature("word_embeddings")
    CHUNK_EMBEDDINGS = JslFeature("chunk_embeddings")
    SENTENCE_EMBEDDINGS = JslFeature("sentence_embeddings")
    CATEGORY = JslFeature("category")
    DATE = JslFeature("date")
    MULTI_DOCUMENT_CLASSIFICATION = JslFeature('multi_document_classification')
    DOCUMENT_CLASSIFICATION = JslFeature('document_classification')
    TOKEN_CLASSIFICATION = JslFeature('token_classification')
    SEQUENCE_CLASSIFICATION = JslFeature('sequence_classification')

    SENTIMENT = JslFeature("sentiment")
    POS = JslFeature("pos")
    CHUNK = JslFeature("chunk")
    NAMED_ENTITY_IOB = JslFeature("named_entity_iob")
    NAMED_ENTITY_CONVERTED = JslFeature("named_entity_converted")
    NAMED_ENTITY_CONVERTED_AND_CONVERTED_TO_DOC = JslFeature("NAMED_ENTITY_CONVERTED_AND_CONVERTED_TO_DOC")
    NEGEX = JslFeature("negex")
    UNLABLED_DEPENDENCY = JslFeature("unlabeled_dependency")
    LABELED_DEPENDENCY = JslFeature("labeled_dependency")
    LANGUAGE = JslFeature("language")
    NODE = JslFeature("node")
    DUMMY = JslFeature("dummy")
    IMAGE = JslFeature("spark_nlp_image")
    CLASSIFIED_IMAGE = JslFeature('classified_image')
    SPARK_NLP_IMAGE = JslFeature('image')
    SPARK_NLP_FILE_PATH = JslFeature('path')

class OCR_FEATURES(JslFeature):
    """
    OCR Features
    """

    BINARY_IMG = JslFeature("content")  # img -
    BINARY_PDF = JslFeature("content")  # pdf bin_pdf
    BINARY_PPT = JslFeature("bin_ppt")  # Powerpoint bin_ppt
    BINARY_PDF_PAGE = JslFeature("bin_pdf_page")  # just a page
    BINARY_DOCX = JslFeature("content")  # pdf2text - bin_docx
    BINARY_DOCX_PAGE = JslFeature("bin_docx_page")  # just a page
    BINARY_TOKEN = JslFeature("bin_token")  # img -
    BINARY_DICOM = JslFeature("bin_dicom")  # DICOM image
    DICOM_METADATA = JslFeature("dicom_metadata")  # DICOM metadata (json formatted)

    FILE_PATH = JslFeature("path")  # TODO this is externalL???
    TEXT = JslFeature("text")  # TODO should be same class as the spark NLP ones TODO EXTERNANMALLL??

    TEXT_ENTITY = JslFeature('text_entity')  # chunk/entity
    TEXT_DOCUMENT = JslFeature("text_document")  # TODO should be same class as the spark NLP ones
    TEXT_DOCUMENT_TOKENIZED = JslFeature("text_tokenized")  # TODO should be same class as the spark NLP ones
    HOCR = JslFeature("hocr")  # img -
    IMG_SPLIT_REGIONS = JslFeature("image_region")
    # All OCR_* features are structs generated from OCR lib
    FALL_BACK = JslFeature("fall_back")  #
    OCR_IMAGE = JslFeature("ocr_image")  # OCR struct image representation
    OCR_PAGE_MATRIX = JslFeature("ocr_page_matrix")  # OCR struct image representation
    OCR_POSITIONS = JslFeature(
        "ocr_positions")  # OCR struct POSITION representation # TODO is POSITIONS==COORDINATES???
    OCR_REGION = JslFeature("region")  # OCR array of POSITION struct
    OCR_TEXT = JslFeature("ocr_text")  # raw text extracted by OCR anno like PDFtoImage
    OCR_TABLE = JslFeature("ocr_table")  # OCR extracted table TODO array of COORDINATES/POSITION?
    OCR_TABLE_CELLS = JslFeature("ocr_table_cells")  # OCR extracted table  TODO array of COORDINATES/POSITION??
    OCR_MAPPING = JslFeature("ocr_table")  # TODO wat is MPAPING???
    PAGE_NUM = JslFeature("page_num")  # TODO is this just int or some struct?

    JSON_FOUNDATION_ONE_REPORT = JslFeature("json_foundation_one_report")

    PREDICTION_TEXT_TABLE = JslFeature("prediction_text_lable")  # TODO is this just int or some struct?
    PREDICTION_CONFIDENCE = JslFeature("prediction_confidence")  # TODO is this just int or some struct?
    VISUAL_CLASSIFIER_CONFIDENCE = JslFeature("visual_classifier_confidence")
    VISUAL_CLASSIFIER_PREDICTION = JslFeature("visual_classifier_prediction")


class NLP_HC_FEATURES(JslFeature):
    """
    NLP HC Feature aka Annotator Types
    """
    ASSERTION = JslFeature('assertion')
    RESOLVED_ENTITY = JslFeature('resolved_entity')
    FEATURE_VECTOR = JslFeature('feature_vector')
    MAPPED_CHUNK = JslFeature('mapped_chunk')
    DISAMBIGUATION = JslFeature('disambiguation')
    RELATION = JslFeature('relation')


"""
Resolve Annotator Classes in the Pipeline to Extractor Configs and Methods

Every Annotator should have 2 configs. Some might offor multuple configs/method pairs, based on model_anno_obj/NLP reference.
- default/minimalistic -> Just the results of the annotations, no confidences or extra metadata
- with meta            -> A config that leverages white/black list and gets the most relevant metadata
- with positions       -> With Begins/Ends
- with sentence references -> Reeturn the sentence/chunk no. reference from the metadata.
                                If a document has multi-sentences, this will map a label back to a corrosponding sentence

"""
from nlu.pipe.extractors.extractor_configs_OS import *
from nlu.pipe.extractors.extractor_configs_HC import *

from sparknlp_jsl.annotator  import *
from sparknlp_jsl.base import *

HC_anno2config = {
    MedicalNerModel : {
        'default': default_ner_config,
        # 'meta': meta_NER_config,
        'default_full'  : default_full_config,
    },
    NerConverterInternal : {
        'default': default_NER_converter_licensed_config,
        'default_full'  : default_full_config,
    },
    AssertionDLModel : {
        'default': default_assertion_config,
        'default_full'  : default_full_config,
    },
    AssertionLogRegModel : {
        'default': default_assertion_config,
        'default_full'  : default_full_config,
    },
    SentenceEntityResolverModel : {
        'default': default_chunk_resolution_config,
        'default_full'  : default_full_config,
    },
    SentenceEntityResolverApproach : {
        'default': default_chunk_resolution_config,
        'default_full'  : default_full_config,
    },
    # ChunkEntityResolverModel : {
    #     'default': default_chunk_resolution_config,
    #     'default_full'  : default_full_config,
    # },
    # ChunkEntityResolverApproach : {
    #     'default': default_chunk_resolution_config,
    #     'default_full'  : default_full_config,
    # },

    DeIdentificationModel : {
        'default': default_de_identification_config,
        'default_full'  : default_full_config,
    },
    RelationExtractionModel : {
        'default': default_relation_extraction_config,
        'positional': default_relation_extraction_positional_config,
        'default_full'  : default_full_config,
    },

    RelationExtractionDLModel : {
        'default': default_relation_extraction_config,
        'positional': default_relation_extraction_positional_config,
        'default_full'  : default_full_config,
    },
    Chunk2Token : {
        'default': '',# TODO
        'default_full'  : default_full_config,
    },

    ContextualParserModel : {
        'default': default_full_config,# TODO
        'default_full'  : default_full_config,

    },

    ContextualParserApproach : {
        'default': default_full_config,# TODO
        'default_full'  : default_full_config,

    },
    DrugNormalizer : {
        'default': default_only_result_config,
        'default_full'  : default_full_config,
    },

    GenericClassifierModel : {
        'default': default_generic_classifier_config,
        'default_full'  : default_full_config,
    },

    GenericClassifierApproach : {
        'default': default_generic_classifier_config,
        'default_full'  : default_full_config,
    },


    FeaturesAssembler : {
        'default': default_feature_assembler_config,
        'default_full'  : default_full_config,
    },


    ChunkMergeModel : {
        'default': '',# TODO
        'default_full'  : default_full_config,
    },

    NerDisambiguatorModel : {
        'default': '',# TODO
        'default_full'  : default_full_config,
    },

    RENerChunksFilter : {
        'default': '',# TODO
        'default_full'  : default_full_config,
    },

    NerOverwriter : {
        'default': '',# TODO
        'default_full'  : default_full_config,
    },
    PosologyREModel : {
        # 'default': '',# TODO
        'default': default_relation_extraction_positional_config,
        'default_full'  : default_full_config,

    }

}










"""Get data into JVM for prediction and out again as Spark Dataframe"""
import logging
from io import StringIO
from typing import Iterable

from nlu.universe.feature_universes import NLP_FEATURES

logger = logging.getLogger('nlu')
import pyspark
from pyspark.sql.functions import monotonically_increasing_id
import numpy as np
import pandas as pd
from pyspark.sql.types import StringType, StructType, StructField


class NluDataParseException(Exception):
    """Custom exception class"""

    def __init__(self, message="An error occurred parsing data with NLU"):
        self.message = message
        super().__init__(self.message)

class DataConversionUtils:
    # Modin aswell but optional, so we dont import the type yet
    supported_types = [pyspark.sql.DataFrame, pd.DataFrame, pd.Series, np.ndarray]

    @staticmethod
    def except_text_col_not_found(cols):
        raise NluDataParseException(
            f'Could not find column named "text" in input Pandas Dataframe. Please ensure one column named such exists. Columns in DF are : {cols} ')

    @staticmethod
    def except_invalid_question_data_format(cols):
        raise NluDataParseException(
            f'You input data format is invalid for question answering with span classification.'
            f'Make sure you have at least 2 columns in you dataset, named context/question  for pandas Dataframes'
            f'For Strings/Iterables/Tuples make sure to use the format `question|||context` or (question,context) ')

    @staticmethod
    def sdf_to_sdf(data, spark_sess, raw_text_column='text'):
        """No casting, Spark to Spark. Just add index col"""
        logger.info(f"Casting Spark DF to Spark DF")
        output_datatype = 'spark'
        data = data.withColumn('origin_index', monotonically_increasing_id().alias('origin_index'))
        stranger_features = []
        if raw_text_column in data.columns:
            # store all stranger features
            if len(data.columns) > 1:
                stranger_features = list(set(data.columns) - set(raw_text_column))
            else:
                DataConversionUtils.except_text_col_not_found(data.columns)
        return data, stranger_features, output_datatype

    @staticmethod
    def table_question_str_to_sdf(data, spark_sess):

        # Its JSON or CSV but as Raw string, validate with json.loads or pd.read_csv()
        try:
            # Validate JSON
            # For now no csv validation because we dont have validator. We let Spark-NLP Pipe validate the date
            return spark_sess.createDataFrame(
                [[data[0], " ".join([[data[1]] if isinstance(data[1], str) else data[1]])]]
            ).toDF(NLP_FEATURES.RAW_QUESTION_CONTEXT, NLP_FEATURES.RAW_QUESTION), [], 'pandas'
        except Exception as err:
            # TODO PARSE FAIl
            raise err

    @staticmethod
    def table_question_pdf_to_sdf(data, spark_sess):
        # For now no csv validation because we dont have delimitor. We let Spark-NLP Pipe validate the date
        # pd.read_csv(StringIO(data[0]))
        header = '"header": ['+",  ".join((list(map(lambda x : f'"{x}"' , data[0].columns))))+ ']'
        json_data_str = '{' +f'''{header}, "rows" : {data[0].to_json(orient='values')} ''' + '}'
        # data[0].to_json()
        try:
            return spark_sess.createDataFrame(
                [[json_data_str, data[1] if isinstance(data[1], str) else ' '.join(data[1])]]
            ).toDF(NLP_FEATURES.RAW_QUESTION_CONTEXT, NLP_FEATURES.RAW_QUESTION).withColumn('origin_index', monotonically_increasing_id().alias('origin_index')), [], 'pandas'
        except Exception as err:
            # TODO PARSE FAIl
            raise err


    @staticmethod
    def question_sdf_to_sdf(data, spark_sess):
        """Casting question pandas to spark and add index col"""
        logger.info(f"Casting Pandas DF to Spark DF")
        output_datatype = 'spark'
        if NLP_FEATURES.RAW_QUESTION not in data.columns or NLP_FEATURES.RAW_QUESTION_CONTEXT not in data.columns:
            if len(data.columns) < 2:
                DataConversionUtils.except_invalid_question_data_format(data)
            data = data.withColumnRenamed(data.columns[0], NLP_FEATURES.RAW_QUESTION) \
                .withColumnRenamed(data.columns[1], NLP_FEATURES.RAW_QUESTION_CONTEXT)

        data = data.withColumn('origin_index', monotonically_increasing_id().alias('origin_index'))
        # make  Nans to None, or spark will crash
        stranger_features = list(set(data.columns) - {NLP_FEATURES.RAW_QUESTION, NLP_FEATURES.RAW_QUESTION_CONTEXT})
        return data, stranger_features, output_datatype


    @staticmethod
    def question_str_to_sdf(data, spark_sess):
        """Casting str  to spark and add index col. This is a bit inefficient. Casting follow  # inefficient, str->pd->spark->pd , we can could first pd"""
        output_datatype = 'string'
        if '|||' not in data:
            DataConversionUtils.except_invalid_question_data_format(data)
        question, context = data.split('|||')
        sdf = spark_sess.createDataFrame(pd.DataFrame({NLP_FEATURES.RAW_QUESTION: question,
                                                       NLP_FEATURES.RAW_QUESTION_CONTEXT: context,
                                                       'origin_index': [0]}, index=[0]))
        return sdf, [], output_datatype


    @staticmethod
    def question_tuple_to_sdf(data, spark_sess):
        """Casting str  to spark and add index col. This is a bit inefficient. Casting follow  # inefficient, str->pd->spark->pd , we can could first pd"""
        output_datatype = 'string'
        question, context = data[0], data[1]
        sdf = spark_sess.createDataFrame(pd.DataFrame({NLP_FEATURES.RAW_QUESTION: question,
                                                       NLP_FEATURES.RAW_QUESTION_CONTEXT: context,
                                                       'origin_index': [0]}, index=[0]))
        return sdf, [], output_datatype


    @staticmethod
    def question_tuple_iterable_to_sdf(data, spark_sess):
        """Casting str  to spark and add index col. This is a bit inefficient. Casting follow  # inefficient, str->pd->spark->pd , we can could first pd"""
        output_datatype = 'string'

        if len(data) == 0:
            DataConversionUtils.except_invalid_question_data_format(data)
        if len(data[0]) != 2:
            DataConversionUtils.except_invalid_question_data_format(data)

        question, context = zip(*[(d[0], d[1]) for d in data])

        sdf = spark_sess.createDataFrame(pd.DataFrame({NLP_FEATURES.RAW_QUESTION: question,
                                                       NLP_FEATURES.RAW_QUESTION_CONTEXT: context,
                                                       'origin_index': [0]}, index=list(range(len(question)))))
        return sdf, [], output_datatype


    @staticmethod
    def question_str_iterable_to_sdf(data, spark_sess):
        """Casting str  to spark and add index col. This is a bit inefficient. Casting follow  # inefficient, str->pd->spark->pd , we can could first pd"""
        output_datatype = 'string'
        if len(data) == 0:
            DataConversionUtils.except_invalid_question_data_format(data)
        if '|||' not in data[0]:
            DataConversionUtils.except_invalid_question_data_format(data)
        question, context = zip(*[d.split('|||') for d in data])
        sdf = spark_sess.createDataFrame(pd.DataFrame({NLP_FEATURES.RAW_QUESTION: question,
                                                       NLP_FEATURES.RAW_QUESTION_CONTEXT: context,
                                                       'origin_index': list(range(len(question)))}))
        return sdf, [], output_datatype


    @staticmethod
    def pdf_to_sdf(data, spark_sess, raw_text_column='text'):
        """Casting pandas to spark and add index col"""
        logger.info(f"Casting Pandas DF to Spark DF")
        output_datatype = 'pandas'
        stranger_features = []
        sdf = None
        # set first col as text column if there is none
        if raw_text_column not in data.columns: data.rename(columns={data.columns[0]: 'text'}, inplace=True)
        data['origin_index'] = data.index
        if raw_text_column in data.columns:
            if len(data.columns) > 1:
                # make  Nans to None, or spark will crash
                data = data.where(pd.notnull(data), None)
                data = data.dropna(axis=1, how='all')
                stranger_features = list(set(data.columns) - set(raw_text_column))
            sdf = spark_sess.createDataFrame(data)
        else:
            DataConversionUtils.except_text_col_not_found(data.columns)
        return sdf, stranger_features, output_datatype


    @staticmethod
    def question_pdf_to_sdf(data, spark_sess):
        """Casting question pandas to spark and add index col"""
        logger.info(f"Casting Pandas DF to Spark DF")
        output_datatype = 'pandas'
        if NLP_FEATURES.RAW_QUESTION not in data.columns or NLP_FEATURES.RAW_QUESTION_CONTEXT not in data.columns:
            if len(data.columns) < 2:
                DataConversionUtils.except_invalid_question_data_format(data)
            data = data.rename(columns={
                data.columns[0]: NLP_FEATURES.RAW_QUESTION,
                data.columns[1]: NLP_FEATURES.RAW_QUESTION_CONTEXT,
            })

        data['origin_index'] = data.index
        # make  Nans to None, or spark will crash
        data = data.where(pd.notnull(data), None)
        data = data.dropna(axis=1, how='all')
        stranger_features = list(set(data.columns) - {NLP_FEATURES.RAW_QUESTION, NLP_FEATURES.RAW_QUESTION_CONTEXT})
        sdf = spark_sess.createDataFrame(data)
        return sdf, stranger_features, output_datatype


    @staticmethod
    def pds_to_sdf(data, spark_sess, raw_text_column='text'):
        """Casting pandas series to spark and add index col.  # for df['text'] colum/series passing casting follows pseries->pdf->spark->pd """
        logger.info(f"Casting Pandas Series to Spark DF")

        output_datatype = 'pandas_series'
        sdf = None
        schema = StructType([StructField(raw_text_column, StringType(), True)])
        data = pd.DataFrame(data).dropna(axis=1, how='all')
        # If series from a column is passed, its column name will be reused.
        if raw_text_column not in data.columns and len(data.columns) == 1:
            data[raw_text_column] = data[data.columns[0]]
        else:
            logger.info(
                f'INFO: NLU will assume {data.columns[0]} as label column since default text column could not be find')
            data[raw_text_column] = data[data.columns[0]]
        data['origin_index'] = data.index
        if raw_text_column in data.columns:
            sdf = spark_sess.createDataFrame(pd.DataFrame(data[raw_text_column]), schema=schema)
        else:
            DataConversionUtils.except_text_col_not_found(data.columns)
        if 'origin_index' not in sdf.columns:
            sdf = sdf.withColumn('origin_index', monotonically_increasing_id().alias('origin_index'))
        return sdf, [], output_datatype


    @staticmethod
    def np_to_sdf(data, spark_sess, raw_text_column='text'):
        """Casting numpy array to spark and add index col. This is a bit inefficient. Casting follow  np->pd->spark->pd. We could cut out the first pd step   """
        logger.info(f"Casting Numpy Array to Spark DF")
        output_datatype = 'numpy_array'
        if len(data.shape) != 1: ValueError(
            f"Exception : Input numpy array must be 1 Dimensional for prediction.. Input data shape is{data.shape}")
        sdf = spark_sess.createDataFrame(pd.DataFrame({raw_text_column: data, 'origin_index': list(range(len(data)))}))
        return sdf, [], output_datatype


    @staticmethod
    def str_to_sdf(data, spark_sess, raw_text_column='text'):
        """Casting str  to spark and add index col. This is a bit inefficient. Casting follow  # inefficient, str->pd->spark->pd , we can could first pd"""
        logger.info(f"Casting String to Spark DF")
        output_datatype = 'string'
        sdf = spark_sess.createDataFrame(pd.DataFrame({raw_text_column: data, 'origin_index': [0]}, index=[0]))
        return sdf, [], output_datatype


    @staticmethod
    def str_list_to_sdf(data, spark_sess, raw_text_column='text'):
        """Casting str list  to spark and add index col. This is a bit inefficient. Casting follow  # # inefficient, list->pd->spark->pd , we can could first pd"""
        logger.info(f"Casting String List to Spark DF")
        output_datatype = 'string_list'
        if all(type(elem) == str for elem in data):
            sdf = spark_sess.createDataFrame(
                pd.DataFrame({raw_text_column: pd.Series(data), 'origin_index': list(range(len(data)))}))
        else:
            ValueError("Exception: Not all elements in input list are of type string.")
        return sdf, [], output_datatype


    @staticmethod
    def fallback_modin_to_sdf(data, spark_sess, raw_text_column='text'):
        """Casting potential Modin data to spark and add index col. # Modin tests, This could crash if Modin not installed """
        logger.info(f"Casting Modin DF to Spark DF")
        sdf = None
        output_datatype = ''
        try:
            import modin.pandas as mpd
            if isinstance(data, mpd.DataFrame):
                data = pd.DataFrame(data.to_dict())  # create pandas to support type inference
                output_datatype = 'modin'
                data['origin_index'] = data.index
            if raw_text_column in data.columns:
                if len(data.columns) > 1:
                    data = data.where(pd.notnull(data), None)  # make  Nans to None, or spark will crash
                    data = data.dropna(axis=1, how='all')
                    stranger_features = list(set(data.columns) - set(raw_text_column))
                sdf = spark_sess.createDataFrame(data)
            else:
                DataConversionUtils.except_text_col_not_found(data.columns)
            if isinstance(data, mpd.Series):
                output_datatype = 'modin_series'
                data = pd.Series(data.to_dict())  # create pandas to support type inference
                data = pd.DataFrame(data).dropna(axis=1, how='all')
                data['origin_index'] = data.index
                index_provided = True
                if raw_text_column in data.columns:
                    sdf = spark_sess.createDataFrame(data[['text']])
                else:
                    DataConversionUtils.except_text_col_not_found(data.columns)
        except:
            print(
                "If you use Modin, make sure you have installed 'pip install modin[ray]' or 'pip install modin[dask]' backend for Modin ")
        return sdf, [], output_datatype


    @staticmethod
    def to_spark_df(data, spark_sess, raw_text_column='text', is_span_data=False, is_tabular_qa_data=False):
        """Convert supported datatypes to SparkDF and extract extra data for prediction later on."""
        if is_tabular_qa_data:
            if not isinstance(data, Iterable):
                # TODO invalid Table Data Format Exception
                pass
            if not isinstance(data[0], (pd.DataFrame, str)):
                # TODO invalid Table Data Format Exception
                pass
            if not isinstance(data[1], Iterable):
                # TODO invalid Table Data Format Exception
                pass
            if isinstance(data[0], str):
                return DataConversionUtils.table_question_str_to_sdf(data, spark_sess)
            if isinstance(data[0], pd.DataFrame):
                return DataConversionUtils.table_question_pdf_to_sdf(data, spark_sess)

        if is_span_data:
            try:
                if isinstance(data, pyspark.sql.dataframe.DataFrame):
                    return DataConversionUtils.question_sdf_to_sdf(data, spark_sess)
                elif isinstance(data, pd.DataFrame):
                    return DataConversionUtils.question_pdf_to_sdf(data, spark_sess)
                elif isinstance(data, tuple):
                    return DataConversionUtils.question_tuple_to_sdf(data, spark_sess)
                elif isinstance(data, str):
                    return DataConversionUtils.question_str_to_sdf(data, spark_sess)
                elif isinstance(data, (list, pd.Series, np.ndarray)):
                    if isinstance(data[0], tuple):
                        return DataConversionUtils.question_tuple_iterable_to_sdf(data, spark_sess)
                    elif isinstance(data[0], str):
                        return DataConversionUtils.question_str_iterable_to_sdf(data, spark_sess)
            except NluDataParseException as err :
                raise err
            except:
                ValueError("Data could not be converted to Spark Dataframe for internal conversion.")
        else:
            try:
                if isinstance(data, pyspark.sql.dataframe.DataFrame):
                    return DataConversionUtils.sdf_to_sdf(data, spark_sess, raw_text_column)
                elif isinstance(data, pd.DataFrame):
                    return DataConversionUtils.pdf_to_sdf(data, spark_sess, raw_text_column)
                elif isinstance(data, pd.Series):
                    return DataConversionUtils.pds_to_sdf(data, spark_sess, raw_text_column)
                elif isinstance(data, np.ndarray):
                    return DataConversionUtils.np_to_sdf(data, spark_sess, raw_text_column)
                elif isinstance(data, str):
                    return DataConversionUtils.str_to_sdf(data, spark_sess, raw_text_column)
                elif isinstance(data, list):
                    return DataConversionUtils.str_list_to_sdf(data, spark_sess, raw_text_column)
                else:
                    return DataConversionUtils.fallback_modin_to_sdf(data, spark_sess, raw_text_column)
            except:
                ValueError("Data could not be converted to Spark Dataframe for internal conversion.")
        raise TypeError(f"Invalid datatype = {type(data)}")


    @staticmethod
    def str_to_pdf(data, raw_text_column):
        logger.info(f"Casting String to Pandas DF")
        return pd.DataFrame({raw_text_column: [data]}).reset_index().rename(
            columns={'index': 'origin_index'}), [], 'string'


    @staticmethod
    def str_list_to_pdf(data, raw_text_column):
        logger.info(f"Casting String List to Pandas DF")
        return pd.DataFrame({raw_text_column: data}).reset_index().rename(
            columns={'index': 'origin_index'}), [], 'string_list'


    @staticmethod
    def np_to_pdf(data, raw_text_column):
        logger.info(f"Casting Numpy Array to Pandas DF")
        return pd.DataFrame({raw_text_column: data}).reset_index().rename(
            columns={'index': 'origin_index'}), [], 'string_list'


    @staticmethod
    def pds_to_pdf(data, raw_text_column):
        return pd.DataFrame({raw_text_column: data}).reset_index().rename(
            columns={'index': 'origin_index'}), [], 'string_list'


    @staticmethod
    def pdf_to_pdf(data, raw_text_column):
        logger.info(f"Casting Pandas DF to Pandas DF")
        data = data.reset_index().rename(columns={'index': 'origin_index'})
        stranger_features = list(data.columns)
        if raw_text_column not in stranger_features:
            print(f"Could not find {raw_text_column} col in df. Using {stranger_features[0]} col istead")
            data = data.reset_index().rename(columns={stranger_features[0]: raw_text_column})
        stranger_features.remove('text')
        stranger_features.remove('origin_index')

        return data, stranger_features, 'pandas'


    @staticmethod
    def sdf_to_pdf(data, raw_text_column):
        logger.info(f"Casting Spark DF to Pandas DF")
        data = data.toPandas().reset_index().rename(columns={'index': 'origin_index'})
        stranger_features = list(data.columns)
        if raw_text_column not in stranger_features:
            print(f"Could not find {raw_text_column} col in df. Using {stranger_features[0]} col istead")
            data = data.reset_index().rename(columns={stranger_features[0]: raw_text_column})
        stranger_features.remove('text')
        stranger_features.remove('origin_index')

        return data, stranger_features, 'spark'


    @staticmethod
    def to_pandas_df(data, raw_text_column='text'):
        """
        Convert data to LihgtPipeline Compatible Format, which is np.array[str], list[str] and str  but we need list anyways later.
        So we create here a pd.Dataframe with a TEXT col if not already given
        Convert supported datatypes to Pandas and extract extra data for prediction later on.

        """
        try:
            if isinstance(data, pyspark.sql.dataframe.DataFrame):
                return DataConversionUtils.pdf_to_pdf(data, raw_text_column)
            elif isinstance(data, pd.DataFrame):
                return DataConversionUtils.pdf_to_pdf(data, raw_text_column)
            elif isinstance(data, pd.Series):
                return DataConversionUtils.pds_to_pdf(data, raw_text_column)
            elif isinstance(data, np.ndarray):
                return DataConversionUtils.np_to_pdf(data, raw_text_column)
            elif isinstance(data, str):
                return DataConversionUtils.str_to_pdf(data, raw_text_column)
            elif isinstance(data, list):
                return DataConversionUtils.str_list_to_pdf(data, raw_text_column)
            else:
                return DataConversionUtils.fallback_modin_to_pdf(data, raw_text_column)
        except:
            ValueError("Data could not be converted to Spark Dataframe for internal conversion.")


    @staticmethod
    def size_of(data):
        """
        Convert data to LihgtPipeline Compatible Format, which is np.array[str], list[str] and str  but we need list anyways later.
        So we create here a pd.Dataframe with a TEXT col if not already given
        Convert supported datatypes to Pandas and extract extra data for prediction later on.

        """
        if isinstance(data, pyspark.sql.dataframe.DataFrame):
            return data.count()
        elif isinstance(data, pd.DataFrame):
            return data.shape[0]
        elif isinstance(data, pd.Series):
            return data.shape[0]
        elif isinstance(data, np.ndarray):
            return data.shape[0]
        elif isinstance(data, str):
            return 1
        elif isinstance(data, list):
            return len(data)
        else:
            return len(data)



import logging
from nlu.pipe.utils.resolution.nlu_ref_utils import *
from nlu.spellbook import Spellbook
from nlu.universe.feature_universes import NLP_FEATURES

logger = logging.getLogger('nlu')


def resolve_storage_ref(lang, storage_ref, missing_component_type):
    """Returns a nlp_ref, nlu_ref and whether it is a licensed model_anno_obj or not and an updated languiage, if multi lingual"""
    logger.info(
        f"Resolving storage_ref={storage_ref} for lang={lang} and missing_component_type={missing_component_type}")
    nlu_ref, nlp_ref, is_licensed = None, None, False
    # get nlu ref

    # check if storage_ref is hardcoded
    if lang in Spellbook.licensed_storage_ref_2_nlu_ref.keys() and storage_ref in \
            Spellbook.licensed_storage_ref_2_nlu_ref[lang].keys():
        nlu_ref = Spellbook.licensed_storage_ref_2_nlu_ref[lang][storage_ref]
        is_licensed = True
    elif lang in Spellbook.storage_ref_2_nlu_ref.keys() and storage_ref in Spellbook.storage_ref_2_nlu_ref[
        lang].keys():
        nlu_ref = Spellbook.storage_ref_2_nlu_ref[lang][
            storage_ref]  # a HC model_anno_obj may use OS storage_ref_provider, so we dont know yet if it is licensed or not
    if lang in Spellbook.pretrained_models_references.keys() and nlu_ref in \
            Spellbook.pretrained_models_references[lang].keys():
        nlp_ref = Spellbook.pretrained_models_references[lang][nlu_ref]
    elif lang in Spellbook.pretrained_healthcare_model_references.keys() and nlu_ref in \
            Spellbook.pretrained_healthcare_model_references[lang].keys():
        nlp_ref = Spellbook.pretrained_healthcare_model_references[lang][nlu_ref]
        is_licensed = True
    # check if storage_ref matches nlu_ref and get NLP_ref
    elif lang in Spellbook.licensed_storage_ref_2_nlu_ref.keys() and storage_ref in \
            Spellbook.licensed_storage_ref_2_nlu_ref[lang].keys():
        nlu_ref = storage_ref
        nlp_ref = Spellbook.licensed_storage_ref_2_nlu_ref[lang][nlu_ref]
    elif lang in Spellbook.pretrained_models_references.keys() and storage_ref in \
            Spellbook.pretrained_models_references[lang].keys():
        nlu_ref = storage_ref
        nlp_ref = Spellbook.pretrained_models_references[lang][nlu_ref]

    # check if storage_ref matches nlp_ref and get nlp and nlu ref
    elif lang in Spellbook.pretrained_healthcare_model_references.keys():
        if storage_ref in Spellbook.pretrained_healthcare_model_references[lang].values():
            inv_namespace = {v: k for k, v in Spellbook.pretrained_healthcare_model_references[lang].items()}
            nlp_ref = storage_ref
            nlu_ref = inv_namespace[nlp_ref]
            is_licensed = True

    if nlu_ref is not None and 'xx.' in nlu_ref: lang = 'xx'

    if nlp_ref is None and nlu_ref is not None:
        # cast NLU ref to NLP ref
        if is_licensed:
            nlp_ref = Spellbook.pretrained_healthcare_model_references[lang][nlu_ref]
        else:
            nlp_ref = Spellbook.pretrained_models_references[lang][nlu_ref]

    if nlp_ref is not None and nlu_ref is None:
        # cast NLP ref to NLU ref
        if is_licensed:
            inv_namespace = {v: k for k, v in Spellbook.pretrained_healthcare_model_references[lang].items()}
            nlu_ref = inv_namespace[nlp_ref]
        else:
            inv_namespace = {v: k for k, v in Spellbook.pretrained_models_references[lang].items()}
            nlu_ref = inv_namespace[nlp_ref]

    if nlu_ref == None and nlp_ref == None:
        # todo enfore storage ref when training
        logger.info(f"COULD NOT RESOLVE STORAGE_REF={storage_ref}")
        if storage_ref == '':
            if missing_component_type == NLP_FEATURES.SENTENCE_EMBEDDINGS:
                logger.info("Using default storage_ref USE, assuming training mode")
                storage_ref = 'en.embed_sentence.use'  # this enables default USE embeds for traianble components
                nlp_ref = 'tfhub_use'
                nlu_ref = storage_ref
            elif missing_component_type == NLP_FEATURES.WORD_EMBEDDINGS:
                logger.info("Using default storage_ref GLOVE, assuming training mode")
                storage_ref = 'en.glove'  # this enables default USE embeds for traianble components
                nlp_ref = 'glove_100d'
                nlu_ref = storage_ref
        else:
            nlp_ref = storage_ref
            nlu_ref = storage_ref
    if nlu_ref is not None:
        is_licensed = check_if_nlu_ref_is_licensed(nlu_ref)

    logger.info(f'Resolved storageref = {storage_ref} to NLU_ref = {nlu_ref} and NLP_ref = {nlp_ref}')
    return nlu_ref, nlp_ref, is_licensed, lang


def set_storage_ref_and_resolution_on_component_info(c, storage_ref):
    """Sets a storage ref on a components component_to_resolve info and returns the component_to_resolve """
    c.storage_ref = storage_ref
    return c



from sparknlp.base import *

class SparkNlpImageAssembler:
    @staticmethod
    def get_default_model():
        return ImageAssembler() \
            .setInputCol("image") \
            .setOutputCol("image_assembler")




class RelationExtraction:
    @staticmethod
    def get_default_model():
        from sparknlp_jsl.annotator import RelationExtractionModel
        return RelationExtractionModel.pretrained() \
            .setInputCols(["word_embeddings", "pos", "chunk", "dependency"]) \
            .setOutputCol("relations")

    @staticmethod
    def get_pretrained_model(name, language, bucket='clinical/models'):
        from sparknlp_jsl.annotator import RelationExtractionModel
        if name == "posology_re": bucket = None
        return RelationExtractionModel.pretrained(name, language, bucket) \
            .setInputCols(["word_embeddings", "pos", "entities", "unlabeled_dependency"]) \
            .setOutputCol("relations")

    @staticmethod
    def get_default_trainable_model():
        from sparknlp_jsl.annotator import RelationExtractionApproach
        return RelationExtractionApproach() \
            .setInputCols(["word_embeddings", "pos", "entities", "unlabeled_dependency"]) \
            .setOutputCol("relations") \
            .setLabelCol('label')



from nlu.pipe.extractors.extractor_base_data_classes import SparkNLPExtractor, SparkNLPExtractorConfig
from nlu.pipe.extractors.extractor_methods.helper_extractor_methods import *
from nlu.pipe.extractors.extractor_methods.base_extractor_methods import *

"""
This file contains methods to get pre-defined configurations for every annotator in OCR
Extractor_resolver.py should be used to resolve SparkNLP Annotator classes to methods 
in this file, which return the corrosponding configs that need to be passed to 
the master_extractor() call.

This file is where all the in extractor_base_data_classes.py Dataclasses are combined with the 
extractors defined in extractor_methods.py.

"""
def default_text_recognizer_config(output_col_prefix='ocr_text'):
    return SparkOCRExtractorConfig(
        get_text=True,
        output_col_prefix=output_col_prefix,
        name='default text recognizer config',
        description='Gets the text from an text recognizing OCR transformer',
    )


def default_visual_classifier_config(output_col_prefix='visual_classifier'):
    return SparkOCRExtractorConfig(
        output_col_prefix=output_col_prefix,
        name='visual_classifier label and confidence ',
        description='Gets label and confidence of visual classifier',
    )


def default_binary_to_image_config(output_col_prefix='binary_image'):
    return SparkOCRExtractorConfig(
        output_col_prefix=output_col_prefix,
        get_image_origin=True,
        get_image_height=True,
        get_image_width=True,
        get_image_n_channels=True,
        get_image_mode=True,
        get_image_resolution=True,
        get_image_data = True,
        name='FULL binary to image extractor ',
        description='Gets all fields generated by binary to image Transformer ',
    )



from sparknlp.annotator import *

class Yake:
    @staticmethod
    def get_default_model():
        return YakeKeywordExtraction() \
                   .setInputCols("token") \
                   .setOutputCol("keywords") \
                   .setMinNGrams(1) \
                   .setMaxNGrams(3) \
                   .setNKeywords(3) \


    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return Yake.get_default_model()





from sparknlp.annotator import *

class SentimentDl:
    @staticmethod
    def get_default_model():
        return SentimentDLModel.pretrained() \
            .setInputCols("sentence_embeddings") \
            .setOutputCol("category") \


    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return SentimentDLModel.pretrained(name,language,bucket) \
            .setInputCols("sentence_embeddings") \
            .setOutputCol("category") \


    @staticmethod
    def get_default_trainable_model():
        return SentimentDLApproach() \
            .setInputCols("sentence_embeddings") \
            .setOutputCol("category") \
            .setLabelColumn("y") \
            .setMaxEpochs(1) \
            .setEnableOutputLogs(True)


__version__ = '5.1.5rc2'


import nlu.utils.environment.env_utils as env_utils

if not env_utils.try_import_pyspark_in_streamlit():
    raise ImportError("You ned to install Pyspark to run nlu. Run pip install pyspark==3.0.1")
if not env_utils.try_import_spark_nlp():
    raise ImportError("You need Spark NLP to run NLU. run pip install spark-nlp")

import sparknlp
import warnings
import nlu.utils.environment.authentication as auth_utils
import nlu.utils.environment.offline_load_utils as offline_utils
from nlu.universe.universes import Licenses
from nlu.utils.environment.authentication import *
from nlu.pipe.pipeline import NLUPipeline
from nlu.pipe.pipe_logic import PipelineCompleter
from nlu.discovery import Discoverer
from nlu.pipe.component_resolution import *

# apply hotfix
import pandas as pd
pd.DataFrame.iteritems = pd.DataFrame.items

def version(): return __version__


warnings.filterwarnings("ignore")
logger = logging.getLogger('nlu')
logger.setLevel(logging.CRITICAL)
ch = logging.StreamHandler()
ch.setLevel(logging.CRITICAL)
logger.addHandler(ch)
st_cache_enabled = False
nlu_package_location = nlu.__file__[:-11]
discoverer = Discoverer()

slack_link = 'https://join.slack.com/t/spark-nlp/shared_invite/zt-lutct9gm-kuUazcyFKhuGY3_0AMkxqA'
github_issues_link = 'https://github.com/JohnSnowLabs/nlu/issues'


def viz(nlp_pipe: Union[Pipeline, LightPipeline, PipelineModel, List],
        data: str, viz_type: str = '',
        labels_to_viz=None,
        viz_colors={},
        return_html=False,
        ner_col: str = None,
        pos_col: str = None,
        dep_untyped_col: str = None,
        dep_typed_col: str = None,
        resolution_col: str = None,
        relation_col: str = None,
        assertion_col: str = None, ):
    """
Visualize input data with an already configured Spark NLP pipeline,
for  Algorithms of type (Ner,Assertion, Relation, Resolution, Dependency)
using Spark-NLP-Display.
Automatically infers applicable viz type and output columns to use for visualization.
Data may only be a string.


If a pipeline has multiple models candidates that can be used for a viz,
the first Annotator that is vizzable will be used to create viz.
You can specify which type of viz to create with the viz_type parameter

Output columns to use for the viz are automatically deducted from the pipeline, by using the
first pipe that provides the correct output type for a specific viz.
You can specify which columns to use for a viz by using the
corresponding ner_col, pos_col, dep_untyped_col, dep_typed_col, resolution_col, relation_col, assertion_col, parameters.

    :param nlp_pipe: One of  [Pipeline, LightPipeline, PipelineModel, List[Annotator] ]
    :param data:  String to viz
    :param viz_type: Viz type, one of [ner,dep,resolution,relation,assert].
            If none defined, nlu will infer and apply all applicable viz
    :param labels_to_viz: Defines a subset of NER labels to viz i.e. ['PER'] ,
        by default=[] which will display all labels. Applicable only for NER viz

    :param viz_colors: Applicable for [ner, resolution, assert ],
        a dictionary where key = label, value=hex color,
        i.e. viz_colors={'TREATMENT':'#008080', 'problem':'#800080'}
    :param return_html:
    :param ner_col: Specify NER column for NER/Resolution/Assertion viz.
    :param pos_col: Specify POS column for Dependency Tree viz
    :param dep_untyped_col: Specify Untyped Dependency Column for Tree Viz
    :param dep_typed_col: Specify Typed Dependency Column for Tree Viz
    :param resolution_col: Specify Resolution col for resolution viz
    :param relation_col: Specify Relation col for Relation viz
    :param assertion_col: Specify Assertion col for Assertion viz
    :return:
    """
    return to_nlu_pipe(nlp_pipe, True).viz(text_to_viz=data,
                                           viz_type=viz_type,
                                           labels_to_viz=labels_to_viz,
                                           return_html=return_html,
                                           viz_colors=viz_colors,
                                           ner_col=ner_col,
                                           pos_col=pos_col,
                                           dep_untyped_col=dep_untyped_col,
                                           dep_typed_col=dep_typed_col,
                                           resolution_col=resolution_col,
                                           relation_col=relation_col,
                                           assertion_col=assertion_col, )


def autocomplete_pipeline(pipe: Union[Pipeline, LightPipeline, PipelineModel, List], lang='en'):
    """
Auto-Complete a pipeline or single annotator into a runnable pipeline by harnessing NLU's DAG Autocompletion algorithm
and returns it as NLU pipeline.
The standard Spark pipeline is available on the `.vanilla_transformer_pipe` attribute of the returned nlu pipe

Every Annotator and Pipeline of Annotators defines a `DAG` of tasks, with various
dependencies that must be satisfied in `topological order`.
NLU enables the completion of an incomplete DAG by finding or creating a path between
the very first input node which is almost always is `DocumentAssembler/MultiDocumentAssembler`
and the very last node(s), which is given by the `topological sorting` the iterable annotators' parameter.
Paths are created by resolving input features of
 annotators to the corresponding providers with matching storage references.




    :param pipe: Pipeline, list of Annotators or single annotator to complete
    :param lang: Language of the elements in pipe, used to resolve correct language for dependencies of the pipe.
    :return: NLU pipeline with completed pipeline
    """
    # If you dont set lang, you can get storage ref errors!
    if isinstance(pipe, List):
        pipe = to_nlu_pipe(pipe, is_pre_configured=False)
    else:
        pipe = to_nlu_pipe([pipe], is_pre_configured=False)
    pipe = PipelineCompleter.check_and_fix_nlu_pipeline(pipe)
    return pipe


def to_pretty_df(nlp_pipe: Union[Pipeline, LightPipeline, PipelineModel, List], data, positions=False,
                 output_level='', metadata=False, ):
    """
Annotates a Pandas Dataframe/Pandas Series/Numpy Array/Spark DataFrame/Python List strings /Python String
with given Spark NLP pipeline, which is assumed to be complete and runnable.

Annotators are grouped internally by NLU into specific output levels, which will be zipped and exploded together to create
the final output df. Additionally, most  keys from the metadata dictionary in the result annotations will be
expanded into their own columns in the resulting Dataframe.
Some columns are omitted from metadata to reduce total amount of output columns, these can be re-enabled by setting metadata=True


    :param nlp_pipe: Pipeline or List of annotators to use for prediction
    :param data: Data to predict on
    :param output_level: output level, either document/sentence/chunk/token/relation
    :param positions: whether to output indexes that map predictions back to position in origin string
    :param metadata: whether to keep additional metadata in final df or not like
            confidences of every possible class for predictions.
    """
    return to_nlu_pipe(nlp_pipe, True).predict(data, positions=positions,
                                               output_level=output_level,
                                               metadata=metadata)


def to_nlu_pipe(nlp_pipe: Union[Pipeline, LightPipeline, PipelineModel, List], is_pre_configured=True) -> NLUPipeline:
    """
    Convert a pipeline or list which contains sparknlp/sparknlp_jsl annotators
    into NLU pipeline, while maintaining original configuration.
    The pipeline does not need to be pre-fitted.
    :param nlp_pipe: the pipeline to convert, must have iterable attribute of pipe stages.
    :param is_pre_configured: Is the pipeline already configured, i.e. input/output cols are properly matched between anotators? \
            If True, NLU will treat this as a already configured pipeline and will not do any pipeline autocompletion or configs
    :return: nlu pipe wrapping
    """
    pipe = NLUPipeline()
    components = get_nlu_pipe_for_nlp_pipe(nlp_pipe, is_pre_configured)
    for c in components:
        pipe.add(c, is_pre_configured)
        if c.license == Licenses.hc:
            pipe.has_licensed_components = True
    return pipe


def load(request: str = 'from_disk', path: Optional[str] = None, verbose: bool = False, gpu: bool = False,
         streamlit_caching: bool = False,
         m1_chip: bool = False
         ) -> NLUPipeline:
    '''
    Load either a prebuild pipeline or a set of components identified by a whitespace seperated list of components
    You must call nlu.auth() BEFORE calling nlu.load() to access licensed models.
    If you did not call nlu.auth() but did call nlu.load() you must RESTART your Python Process and call nlu.auth().
    You cannot authorize once nlu.load() is called because of Spark Context.
    :param verbose: Whether to output debug prints
    :param gpu: Whether to leverage GPU
    :param streamlit_caching: Whether streamlit caching should be used in Streamlit visualizations. Trade Speed-Up for repeated requests for larger memory usage
    :param path: If path is not None, the model_anno_obj/component_list for the NLU reference will be loaded from the path. Useful for offline mode. Currently only loading entire NLU pipelines is supported, but not loading singular pipes
    :param request: A NLU model_anno_obj/pipeline/component_to_resolve reference. You can request multiple components by separating them with whitespace. I.e. nlu.load('elmo bert albert')
    :return: returns a non fitted nlu pipeline object
    '''
    if streamlit_caching and not nlu.st_cache_enabled:
        enable_streamlit_caching()
        return nlu.load(request, path, verbose, gpu, streamlit_caching)
    # check if secrets are in default loc, if yes load them and create licensed context automatically
    auth(gpu=gpu)
    spark = get_open_source_spark_context(gpu, m1_chip)
    # spark.catalog.clearCache()

    if verbose:
        enable_verbose()
    else:

        disable_verbose()
    try:
        if path is not None:
            logger.info(f'Trying to load nlu pipeline from local hard drive, located at {path}')
            pipe = load_nlu_pipe_from_hdd(path, request)
            return pipe
    except Exception as err:
        if verbose:
            log_verbose_error(err)
        raise Exception(
            f"Something while loading the pipe in {path}. Is the path correct? use nlu.load(verbose=True) for more info.")

    # Try to manifest SparkNLP Annotator from nlu_ref
    components_requested = request.split(' ')
    pipe = NLUPipeline()
    language = parse_language_from_nlu_ref(request)
    pipe.lang = language
    pipe.nlu_ref = request
    try:
        for nlu_ref in components_requested:
            # Iterate over each nlu_ref in the request. Multiple nlu_refs can be passed by seperating them via whitesapce
            nlu_ref.replace(' ', '')
            if nlu_ref == '':
                continue
            nlu_component = nlu_ref_to_component(nlu_ref)
            # if we get a list of components, then the NLU reference is a pipeline, we do not need to check order
            if isinstance(nlu_component, list):
                # lists are parsed down to multiple components, result of pipeline request (stack of components)
                for c in nlu_component:
                    pipe.add(c, nlu_ref, pretrained_pipe_component=True)
            else:
                # just a single component_to_resolve requested
                pipe.add(nlu_component, nlu_ref)
    except Exception as err:
        if verbose:
            log_verbose_error(err)
        raise Exception(
            f"Something went wrong during creating the Spark NLP model_anno_obj for your request = {request}"
            f"Did you use a NLU Spell?")
    # Complete Spark NLP Pipeline, which is defined as a DAG given by the starting Annotators
    try:
        pipe = PipelineCompleter.check_and_fix_nlu_pipeline(pipe)
        pipe.nlu_ref = request
        return pipe
    except:
        if verbose:
            e = sys.exc_info()
            print(e[0])
            print(e[1])
        raise Exception(f"Something went wrong during completing the DAG for the Spark NLP Pipeline."
                        f"If this error persists, please contact us in Slack {slack_link} "
                        f"Or open an issue on Github {github_issues_link}")


def auth(HEALTHCARE_LICENSE_OR_JSON_PATH='/content/spark_nlp_for_healthcare.json', AWS_ACCESS_KEY_ID='',
         AWS_SECRET_ACCESS_KEY='', HEALTHCARE_SECRET='', OCR_LICENSE='', OCR_SECRET='', gpu=False):
    """ Authenticate environment for JSL Licensed models
    Installs NLP-Healthcare if not in environment detected
    Either provide path to spark_nlp_for_healthcare.json file as first param or manually enter them,
    HEALTHCARE_LICENSE_OR_JSON_PATH,AWS_ACCESS_KEY_ID,AWS_SECRET_ACCESS_KEY,HEALTHCARE_SECRET .
    Set gpu=true if you want to enable GPU mode
    """
    auth_utils.auth(HEALTHCARE_LICENSE_OR_JSON_PATH=HEALTHCARE_LICENSE_OR_JSON_PATH,
                    AWS_SECRET_ACCESS_KEY=AWS_SECRET_ACCESS_KEY,
                    AWS_ACCESS_KEY_ID=AWS_ACCESS_KEY_ID, HEALTHCARE_SECRET=HEALTHCARE_SECRET,
                    OCR_LICENSE=OCR_LICENSE,
                    OCR_SECRET=OCR_SECRET, gpu=gpu)
    return nlu


def is_nlu_uid(uid: str):
    return 'is_nlu_pipe' in uid


def load_nlu_pipe_from_hdd_in_databricks(pipe_path, request) -> NLUPipeline:
    """Either there is a pipeline of models in the path or just one singular model_anno_obj.
    If it is a component_list,  load the component_list and return it.
    If it is a singular model_anno_obj, load it to the correct AnnotatorClass and NLU component_to_resolve and then generate pipeline for it
    """

    # def dbfs_path_exist(path):
    #     try:
    #         dbutils.fs.ls(path)
    #         return True
    #     except:
    #         return False
    from pyspark.dbutils import DBUtils
    dbutils = DBUtils(sparknlp.start())
    dbutils.fs

    is_pipe = lambda path: any([f.name == 'stages/' for f in dbutils.fs.ls(path)])
    is_model = lambda path: any([f.name == 'metadata/' for f in dbutils.fs.ls(path)])

    pipe = NLUPipeline()
    nlu_ref = request  # pipe_path
    # if dbfs_path_exist(pipe_path):
    # Resource in path is a pipeline
    if is_pipe(pipe_path):
        pipe_components, uid = get_trained_component_list_for_nlp_pipe_ref('en', nlu_ref, nlu_ref, pipe_path, False)
    # Resource in path is a single model_anno_obj
    elif is_model(pipe_path):
        c = offline_utils.verify_and_create_model(pipe_path)
        c.nlu_ref = nlu_ref
        pipe.add(c, nlu_ref, pretrained_pipe_component=True)
        return PipelineCompleter.check_and_fix_nlu_pipeline(pipe)

    else:
        # fallback pipe
        pipe_components, uid = get_trained_component_list_for_nlp_pipe_ref('en', nlu_ref, nlu_ref, pipe_path, False)
    for c in pipe_components: pipe.add(c, nlu_ref, pretrained_pipe_component=True)
    return pipe


def load_nlu_pipe_from_hdd(pipe_path, request) -> NLUPipeline:
    """Either there is a pipeline of models in the path or just one singular model_anno_obj.
    If it is a component_list,  load the component_list and return it.
    If it is a singular model_anno_obj, load it to the correct AnnotatorClass and NLU component_to_resolve and then generate pipeline for it
    """
    if is_running_in_databricks():
        return load_nlu_pipe_from_hdd_in_databricks(pipe_path, request)
    pipe = NLUPipeline()
    pipe.nlu_ref = request
    nlu_ref = request  # pipe_path
    if os.path.exists(pipe_path):

        # Resource in path is a pipeline
        if offline_utils.is_pipe(pipe_path):
            # language, nlp_ref, nlu_ref,path=None, is_licensed=False
            # todo deduct lang and if Licensed or not
            pipe_components, uid = get_trained_component_list_for_nlp_pipe_ref('en', nlu_ref, nlu_ref, pipe_path, False)
        # Resource in path is a single model_anno_obj
        elif offline_utils.is_model(pipe_path):
            c = offline_utils.verify_and_create_model(pipe_path)
            c.nlu_ref = nlu_ref
            pipe.add(c, nlu_ref, pretrained_pipe_component=True)
            return PipelineCompleter.check_and_fix_nlu_pipeline(pipe)
        else:
            print(
                f"Could not load model_anno_obj in path {pipe_path}. Make sure the jsl_folder contains either a stages subfolder or a metadata subfolder.")
            raise ValueError
        for c in pipe_components:
            pipe.add(c, nlu_ref, pretrained_pipe_component=True)
        if is_nlu_uid(uid):
            data = json.loads(uid)
            print(data)
            pipe.nlu_ref = data['0']['nlu_ref']
            for i, c in enumerate(pipe.components):
                c.nlu_ref = data[str(i + 1)]['nlu_ref']
                c.nlp_ref = data[str(i + 1)]['nlp_ref']
                c.loaded_from_pretrained_pipe = data[str(i + 1)]['loaded_from_pretrained_pipe']
        return pipe

    else:
        print(
            f"Could not load model_anno_obj in path {pipe_path}. Make sure the jsl_folder contains either a stages subfolder or a metadata subfolder.")
        raise ValueError


def get_open_source_spark_context(gpu, m1_chip):
    if env_utils.is_env_pyspark_3_x():
        if m1_chip:
            return sparknlp.start(gpu=gpu, m1=True)
        else:
            return sparknlp.start(gpu=gpu)
    raise ValueError(f"Failure starting Spark Context! Current Spark version {get_pyspark_version()} not supported! "
                     f"Please install any of Pyspark 3.X versions.")


def enable_verbose() -> None:
    logger.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    logger.addHandler(ch)


def disable_verbose() -> None:
    logger.setLevel(logging.ERROR)
    ch = logging.StreamHandler()
    ch.setLevel(logging.ERROR)
    logger.addHandler(ch)


def enable_streamlit_caching():
    # dynamically monkeypatch aka replace the nlu.load() method with the wrapped st.cache if streamlit_caching
    nlu.st_cache_enabled = True
    nlu.non_caching_load = load
    nlu.load = wrap_with_st_cache_if_available_and_set_layout_to_wide(nlu.load)


# def disable_streamlit_caching(): # WIP not working
#     if hasattr(nlu, 'non_caching_load') : nlu.load = nlu.non_caching_load
#     else : print("Could not disable caching.")


def wrap_with_st_cache_if_available_and_set_layout_to_wide(f):
    """Wrap function with ST cache method if streamlit is importable"""
    try:
        import streamlit as st
        st.set_page_config(layout='wide')
        logger.info("Using streamlit cache for load")
        return st.cache(f, allow_output_mutation=True, show_spinner=False)
    except:
        logger.exception("Could not import streamlit and apply caching")
        print("You need streamlit to run use this method")
        return f


# Discovery
def print_all_languages():
    ''' Print all languages which are available in NLU Spark NLP pointer '''
    discoverer.print_all_languages()


def print_all_nlu_components_for_lang(lang='en', c_type='classifier'):
    '''Print all NLU components available for a language Spark NLP pointer'''
    discoverer.print_all_nlu_components_for_lang(lang, c_type)


def print_components(lang='', action=''):
    '''Print every single NLU reference for models and pipeliens and their Spark NLP pointer
    :param lang: Language requirements for the components filterd. See nlu.languages() for supported languages
    :param action: Components that will be filterd.'''
    discoverer.print_components(lang, action)


def print_component_types():
    ''' Prints all unique component_to_resolve types in NLU'''
    discoverer.print_component_types()


def print_all_model_kinds_for_action(action):
    discoverer.print_all_model_kinds_for_action(action)


def print_all_model_kinds_for_action_and_lang(lang, action):
    discoverer.print_all_model_kinds_for_action_and_lang(lang, action)


def print_trainable_components():
    '''Print every trainable Algorithm/Model'''
    discoverer.print_trainable_components()


def get_components(m_type='', include_pipes=False, lang='', licensed=False, get_all=False):
    return discoverer.get_components(m_type, include_pipes, lang, licensed, get_all)


def log_verbose_error(err):
    import traceback
    print(traceback.format_exc())
    print(err)



from sparknlp.annotator import GPT2Transformer

class GPT2:
    @staticmethod
    def get_default_model():
        return GPT2Transformer.pretrained() \
            .setInputCols("document") \
            .setOutputCol("gpt2")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return GPT2Transformer.pretrained(name, language) \
            .setInputCols("document") \
            .setOutputCol("gpt2")






"""
Collection of universes shared across all libraries (NLP/HC/OCR), which are collections of atoms
"""
from nlu.universe.atoms import LicenseType, ComponentBackend, ComputeContext, ModelBucketType


class Licenses:
    """Definition of licenses"""
    ocr = LicenseType('ocr')
    hc = LicenseType('healthcare')
    open_source = LicenseType('open_source')


class ModelBuckets:
    """Definition of licenses"""
    ocr = ModelBucketType('clinical/ocr')
    hc = ModelBucketType("clinical/models")
    open_source = None #  ModelBucketType(None)


def license_to_bucket(license_to_resole: LicenseType) -> ModelBucketType:
    if license_to_resole == Licenses.open_source:
        return ModelBuckets.open_source
    if license_to_resole == Licenses.hc:
        return ModelBuckets.hc
    if license_to_resole == Licenses.ocr:
        return ModelBuckets.ocr
    return None


class ComponentBackends:
    """Definition of various Component providers/backends"""
    ocr = ComponentBackend('spark ocr')
    hc = ComponentBackend('spark hc')
    open_source = ComponentBackend('spark nlp')
    external = ComponentBackend('external')


class ComputeContexts:
    """Computation contexts which describe where which framework is used to compute results """
    spark = ComputeContext('spark')
    pandas = ComputeContext('pandas')
    numpy = ComputeContext('numpy')
    modin = ComputeContext('modin')
    py_arrow = ComputeContext('py_arrow')



import streamlit as st
def _set_block_container_style(
        max_width: int = 1200,
        max_width_100_percent: bool = True,
        set_colors : bool = False ,
        COLOR : str = 'blue',
        BACKGROUND_COLOR : str = 'white',
):
    # .reportview-container .main .block-container
    if max_width_100_percent:
        max_width_str = f"max-width: 100%;"
    else:
        max_width_str = f"max-width: {max_width}px;"
    if set_colors:
        st.markdown(
            f"""
    <style>
        .reportview-container .main .block-container{{
            {max_width_str}

        }}
        
     .element-container div:nth-child(1) {{
              margin: auto;

        }}
        
        .reportview-container .main {{
            color: {COLOR};
            background-color: {BACKGROUND_COLOR};
        }}
        
    </style>
    """,
            unsafe_allow_html=True,
        )
    else :
        st.markdown(
            f"""
<style>
    .reportview-container .main .block-container{{
        {max_width_str}
          margin: auto;

    }}
     .element-container div:nth-child(1) {{
              margin: auto;

        }}
        
    textarea {{
        background-color:green;
    }}
</style>
""",
            unsafe_allow_html=True,
        )



from sparknlp.base import *


class Doc_2_Chunk:
    @staticmethod
    def get_default_model():
        return Doc2Chunk() \
            .setInputCols("sentence") \
            .setOutputCol("entities")






from sparknlp.annotator import XlmRoBertaEmbeddings

class XLM:
    @staticmethod
    def get_default_model():
        return XlmRoBertaEmbeddings.pretrained() \
        .setInputCols("sentence", "token") \
        .setOutputCol("xlm_roberta")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return XlmRoBertaEmbeddings.pretrained(name, language) \
            .setInputCols("sentence", "token") \
            .setOutputCol("xlm_roberta")






from sparknlp.base import MultiDocumentAssembler

class SparkNlpMultiDocumentAssembler:
    @staticmethod
    def get_default_model():
        return MultiDocumentAssembler() \
            .setInputCols(["question", "context"]) \
            .setOutputCols(["document_question", "document_context"])



from sparknlp.annotator import *


class BertSentenceChunkEmbeds:
    @staticmethod
    def get_default_model():
        from sparknlp_jsl.annotator import BertSentenceChunkEmbeddings
        return BertSentenceChunkEmbeddings.pretrained() \
        .setInputCols("sentence", "token") \
        .setOutputCol("bert")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        from sparknlp_jsl.annotator import BertSentenceChunkEmbeddings
        return BertSentenceChunkEmbeddings.pretrained(name,language,bucket) \
            .setInputCols("sentence", "token") \
            .setOutputCol("bert")







class NerToChunkConverterLicensed:
    @staticmethod
    def get_default_model():
        from sparknlp_jsl.annotator import NerConverterInternal
        return NerConverterInternal() \
            .setInputCols(["sentence", "token", "ner"]) \
            .setOutputCol("entities") 



import os, sys
import logging

logger = logging.getLogger('nlu')


def get_pyspark_version():
    import pyspark
    return pyspark.version.__version__


def get_pyspark_major_and_minor():
    import pyspark
    return pyspark.version.__version__.split('.')[:2]


def is_env_pyspark_3_1():
    v = get_pyspark_major_and_minor()
    if v[0] == '3' and v[1] == '1':
        return True
    return False

def is_env_pyspark_3_2():
    v = get_pyspark_major_and_minor()
    if v[0] == '3' and v[1] == '2':
        return True
    return False


def is_env_pyspark_3_0():
    v = get_pyspark_major_and_minor()
    if v[0] == '3' and v[1] == '0':
        return True
    return False


def is_env_pyspark_2_4():
    v = get_pyspark_major_and_minor()
    if v[0] == '2' and v[1] == '4':
        return True
    return False


def is_env_pyspark_2_3():
    v = get_pyspark_major_and_minor()
    if v[0] == '2' and v[1] == '3':
        return True
    return False

def is_env_pyspark_2_x():
    v = get_pyspark_major_and_minor()
    return int(v[0]) <= 2

def is_env_pyspark_3_x():
    v = get_pyspark_major_and_minor()
    return int(v[0]) == 3




def check_pyspark_install():
    try:
        from pyspark.sql import SparkSession
        try:
            import sparknlp
            v = sparknlp.start().version
            spark_major = int(v.split('.')[0])
            if spark_major >= 3:
                raise Exception()
        except:
            print(
                f"Detected pyspark version={v} Which is >=3.X\nPlease run '!pip install pyspark==2.4.7' or install any pyspark>=2.4.0 and pyspark<3")
            return False
    except:
        print(
            "No Pyspark installed!\nPlease run '!pip install pyspark==2.4.7' or install any pyspark>=2.4.0 with pyspark<3")
        return False
    return True

def try_import_streamlit():
    try:
        import streamlit as st
    except  ImportError:
        print("You need to install Streamlit to run this functionality.")


def is_running_in_databricks():
    """ Check if the currently running Python Process is running in Databricks or not
     If any Environment Variable name contains 'DATABRICKS' this will return True, otherwise False"""
    for k in os.environ.keys():
        if 'DATABRICKS' in k:
            return True
    return False


def install_and_import_package(pkg_name, version='', import_name=''):
    """ Install Spark-NLP-Healthcare PyPI Package in current environment if it cannot be imported and license
    provided """
    import importlib
    try:
        if import_name == '':
            importlib.import_module(pkg_name)
        else:
            importlib.import_module(import_name)
    except ImportError:
        import pip
        if version == '':
            logger.info(f"{pkg_name} could not be imported. Running 'pip install {pkg_name}'...")
        else:
            logger.info(f"{pkg_name} could not be imported. Running 'pip install {pkg_name}=={version}'...")
        pip_major_version = int(pip.__version__.split('.')[0])
        if pip_major_version in [10, 18, 19, 20]:
            # for these versions pip module does not support installing, we install via OS command straight into pip
            # module
            py_path = sys.executable
            if version == '':
                os.system(f'{py_path} -m pip install {pkg_name}')
            else:
                os.system(f'{py_path} -m pip install {pkg_name}=={version}')
        else:
            if version == '':
                pip.main(['install', f'{pkg_name}'])
            else:
                pip.main(['install', f'{pkg_name}=={version}'])
    finally:
        import site
        from importlib import reload
        reload(site)
        # import name is not always the same name as pkg_name we want to import, so it must be specified via import name
        if import_name != '':
            globals()[import_name] = importlib.import_module(import_name)
        else:
            globals()[pkg_name] = importlib.import_module(pkg_name)



def try_import_pyspark_in_streamlit():
    """Try importing Pyspark or display warn message in streamlit"""
    try:
        import pyspark
        from pyspark.sql import SparkSession
    except:
        print("You need Pyspark installed to run NLU. Run <pip install pyspark==3.0.2>")
        try:
            import streamlit as st
            st.error(
                "You need Pyspark, Sklearn, Pyplot, Pandas, Numpy installed to run this app. Run <pip install pyspark==3.0.2 sklearn pyplot numpy pandas>")
        except:
            return False
        return False
    return True

def try_import_spark_nlp():
    """Try importing Spark NLP"""
    try:
        import sparknlp
    except:
        print("You need Spark NLP to run NLU. run pip install spark-nlp")
        return False
    return True


import nlu
from nlu.discovery import Discoverer
from nlu.pipe.utils.resolution.storage_ref_utils import StorageRefUtils
from typing import List, Optional
import streamlit as st
import numpy as np
import pandas as pd
from nlu.pipe.viz.streamlit_viz.styles import _set_block_container_style
from nlu.pipe.viz.streamlit_viz.streamlit_viz_tracker import StreamlitVizTracker
from nlu.pipe.viz.streamlit_viz.viz_building_blocks.block_utils.entity_manifold_utils import EntityManifoldUtils


class EntityEmbeddingManifoldStreamlitBlock():
    @staticmethod
    def viz_streamlit_entity_embed_manifold(
            pipe,  # nlu component_list
            default_texts: List[str] = ("Donald Trump likes to visit New York", "Angela Merkel likes to visit Berlin!", 'Peter hates visiting Paris'),
            title: Optional[str] = "Lower dimensional Manifold visualization for Entity embeddings",
            sub_title: Optional[str] = "Apply any of the 10+ `Manifold` or `Matrix Decomposition` algorithms to reduce the dimensionality of `Entity Embeddings` to `1-D`, `2-D` and `3-D` ",
            default_algos_to_apply: List[str] = ("TSNE", "PCA"),
            target_dimensions: List[int] = (1, 2, 3),
            show_algo_select: bool = True,
            set_wide_layout_CSS: bool = True,
            num_cols: int = 3,
            model_select_position: str = 'side',  # side or main
            key: str = "NLU_streamlit",
            show_infos: bool = True,
            show_logo: bool = True,
            n_jobs: Optional[int] = 3,  # False
    ):

        from nlu.pipe.viz.streamlit_viz.streamlit_utils_OS import StreamlitUtilsOS
        StreamlitVizTracker.footer_displayed = False

        try:
            import plotly.express as px
            from sklearn.metrics.pairwise import distance_metrics
        except:
            st.error(
                "You need the sklearn and plotly package in your Python environment installed for similarity visualizations. Run <pip install sklearn plotly>")

        if show_logo: StreamlitVizTracker.show_logo()
        if set_wide_layout_CSS: _set_block_container_style()
        if title: st.header(title)
        if sub_title: st.subheader(sub_title)
        # if show_logo :VizUtilsStreamlitOS.show_logo()
        # VizUtilsStreamlitOS.loaded_word_embeding_pipes = []
        if isinstance(default_texts, list) : default_texts = '\n'.join(default_texts)
        data = st.text_area('Enter N texts, seperated by new lines to visualize Sentence Embeddings for ',
                            default_texts).split('\n')
        output_level = 'chunk'
        ner_emebed_pipe_algo_selection = []
        loaded_ner_embed_nlu_refs = []
        algos = ['TSNE']
        # A component_list should have a NER and a Word Embedding
        if pipe not in StreamlitVizTracker.loaded_ner_word_embeding_pipes: StreamlitVizTracker.loaded_ner_word_embeding_pipes.append(
            pipe)
        if pipe not in StreamlitVizTracker.loaded_word_embeding_pipes: StreamlitVizTracker.loaded_word_embeding_pipes.append(
            pipe)

        if show_algo_select:
            # Manifold Selection
            exp = st.expander("Select additional manifold and dimension reduction techniques to apply")
            algos = exp.multiselect(
                "Reduce embedding dimensionality to something visualizable",
                options=(
                    "TSNE", "ISOMAP", 'LLE', 'Spectral Embedding', 'MDS', 'PCA', 'SVD aka LSA', 'DictionaryLearning',
                    'FactorAnalysis', 'FastICA', 'KernelPCA', 'LatentDirichletAllocation'),
                default=default_algos_to_apply, )
            ner_emb_components_usable = [e for e in Discoverer.get_components('ner', True, include_aliases=True) if
                                         'embed' not in e and 'sentence' not in e]

            # Find nlu_ref of currenlty loaded component_list
            for p in StreamlitVizTracker.loaded_ner_word_embeding_pipes:
                loaded_ner_embed_nlu_refs.append(p.nlu_ref)

            # NER Selection
            if model_select_position == 'side':
                ner_emebed_pipe_algo_selection = st.sidebar.multiselect(
                    "Pick additional NER Models for the Dimension Reduction", options=ner_emb_components_usable,
                    default=loaded_ner_embed_nlu_refs, key=key)
            else:
                ner_emebed_pipe_algo_selection = exp.multiselect(
                    "Pick additional NER Models for the Dimension Reduction", options=ner_emb_components_usable,
                    default=loaded_ner_embed_nlu_refs, key=key)

        for ner_nlu_ref in ner_emebed_pipe_algo_selection:
            load = True
            for ner_p in StreamlitVizTracker.loaded_ner_word_embeding_pipes:
                if ner_p.nlu_ref == ner_nlu_ref:
                    load = False
                    break
            if not load: continue
            p = nlu.load(ner_nlu_ref)
            if p not in StreamlitVizTracker.loaded_ner_word_embeding_pipes: StreamlitVizTracker.loaded_ner_word_embeding_pipes.append(
                p)
            if p not in StreamlitVizTracker.loaded_word_embeding_pipes: StreamlitVizTracker.loaded_word_embeding_pipes.append(
                p)

        col_index = 0
        cols = st.columns(num_cols)

        def are_cols_full():
            return col_index == num_cols

        for p in StreamlitVizTracker.loaded_ner_word_embeding_pipes:
            p = EntityManifoldUtils.insert_chunk_embedder_to_pipe_if_missing(p)
            predictions = p.predict(data, metadata=True, output_level=output_level, multithread=False).dropna()
            entity_cols = EntityManifoldUtils.get_ner_cols(predictions)
            chunk_embed_col = EntityManifoldUtils.find_chunk_embed_col(predictions)

            # TODO get cols for non default NER? or multi ner setups?
            # features = predictions[EntityManifoldUtils.get_ner_cols(predictions)]
            # e_col = StreamlitUtilsOS.find_embed_col(predictions)
            e_com = StreamlitUtilsOS.find_embed_component(p)
            e_com_storage_ref = StorageRefUtils.extract_storage_ref(e_com)
            emb = predictions[chunk_embed_col]
            mat = np.array([x for x in emb])
            # for ner_emb_p in ps:
            for algo in algos:
                # Only pos values for latent Dirchlet
                if algo == 'LatentDirichletAllocation': mat = np.square(mat)
                if len(mat.shape) > 2: mat = mat.reshape(len(emb), mat.shape[-1])
                hover_data = entity_cols + ['text']
                # calc reduced dimensionality with every algo
                feature_to_color_by = entity_cols[0]
                if 1 in target_dimensions:
                    low_dim_data = StreamlitUtilsOS.get_manifold_algo(algo, 1, n_jobs).fit_transform(mat)
                    x = low_dim_data[:, 0]
                    y = np.zeros(low_dim_data[:, 0].shape)

                    # predictions['text'] = original_text
                    tsne_df = pd.DataFrame({**{'x': x, 'y': y},
                                            **{k: predictions[k] for k in entity_cols},
                                            **{'text': predictions[entity_cols[-1]]}
                                            })
                    fig = px.scatter(tsne_df, x="x", y="y", color=feature_to_color_by, hover_data=hover_data)
                    subh = f"""Word-Embeddings =`{e_com_storage_ref}`, NER-Model =`{p.nlu_ref}`, Manifold-Algo =`{algo}` for `D=1`"""
                    cols[col_index].markdown(subh)
                    cols[col_index].write(fig, key=key)
                    col_index += 1
                    if are_cols_full():
                        cols = st.columns(num_cols)
                        col_index = 0
                if 2 in target_dimensions:
                    low_dim_data = StreamlitUtilsOS.get_manifold_algo(algo, 2, n_jobs).fit_transform(mat)
                    x = low_dim_data[:, 0]
                    y = low_dim_data[:, 1]
                    tsne_df = pd.DataFrame({**{'x': x, 'y': y},
                                            **{k: predictions[k] for k in entity_cols},
                                            **{'text': predictions[entity_cols[-1]]}
                                            })
                    fig = px.scatter(tsne_df, x="x", y="y", color=feature_to_color_by, hover_data=hover_data)
                    subh = f"""Word-Embeddings =`{e_com_storage_ref}`, NER-Model =`{p.nlu_ref}`, Manifold-Algo =`{algo}` for `D=2`"""
                    cols[col_index].markdown(subh)
                    cols[col_index].write(fig, key=key)
                    col_index += 1
                    if are_cols_full():
                        cols = st.columns(num_cols)
                        col_index = 0
                if 3 in target_dimensions:
                    low_dim_data = StreamlitUtilsOS.get_manifold_algo(algo, 3, n_jobs).fit_transform(mat)
                    x = low_dim_data[:, 0]
                    y = low_dim_data[:, 1]
                    z = low_dim_data[:, 2]
                    tsne_df = pd.DataFrame({**{'x': x, 'y': y, 'z': z},
                                            **{k: predictions[k] for k in entity_cols},
                                            **{'text': predictions[entity_cols[-1]]}
                                            })
                    fig = px.scatter_3d(tsne_df, x="x", y="y", z='z', color=feature_to_color_by, hover_data=hover_data)
                    subh = f"""Word-Embeddings =`{e_com_storage_ref}`, NER-Model =`{p.nlu_ref}`, Manifold-Algo =`{algo}` for `D=3`"""
                    cols[col_index].markdown(subh)
                    cols[col_index].write(fig, key=key)
                    col_index += 1
                    if are_cols_full():
                        cols = st.columns(num_cols)
                        col_index = 0

                # Todo fancy embed infos etc
                # if display_embed_information: display_embed_vetor_information(e_com,mat)

            # if display_embed_information:
            #     exp = st.expander("Embedding vector information")
            #     exp.write(embed_vector_info)

        if show_infos:
            # VizUtilsStreamlitOS.display_infos()
            StreamlitVizTracker.display_model_info(pipe.nlu_ref, pipes=[pipe])
            StreamlitVizTracker.display_footer()



import logging
from typing import Dict, Any, List

from nlu.pipe.col_substitution.col_name_substitution_utils import ColSubstitutionUtils
from nlu.universe.atoms import NlpLevel
from nlu.universe.feature_node_ids import NLP_HC_NODE_IDS, NLP_NODE_IDS
from nlu.universe.feature_universes import NLP_FEATURES
from nlu.universe.logic_universes import NLP_LEVELS
from nlu.universe.universes import Licenses

logger = logging.getLogger('nlu')


class OutputLevelUtils:
    """Resolve output level of pipeline and components"""

    @staticmethod
    def infer_prediction_output_level(pipe) -> NlpLevel:
        """
        This function checks the LAST  component_to_resolve of the NLU pipeline and infers
         from that the output level via checking the components' info.
        :param pipe: to infer output level for
        :return returns inferred output level
        """
        # Loop in reverse over component_list and get first output level
        # (???) of non util/sentence_detector/tokenizer/doc_assembler.
        for c in pipe.components[::-1]:
            return OutputLevelUtils.resolve_component_to_output_level(pipe, c)
        # fallback
        return NLP_LEVELS.DOCUMENT

    @staticmethod
    def resolve_input_dependent_component_to_output_level(pipe, component_to_resolve) -> NlpLevel:
        """
        For a given NLU component  which is input dependent , resolve its output level by checking if it's input stem
        from document or sentence based annotators
        :param pipe: the pipeline containing all components
        :param component_to_resolve:  Input dependent component for which we want to know the output level
        :return: output-level of component
        """
        # (1.) A classifier, which is using sentence/document. We just check input cols
        if NLP_FEATURES.DOCUMENT in component_to_resolve.spark_input_column_names:
            return NLP_LEVELS.DOCUMENT
        if NLP_FEATURES.SENTENCE in component_to_resolve.spark_input_column_names:
            return NLP_LEVELS.SENTENCE
        if NLP_FEATURES.DOCUMENT_QUESTION_CONTEXT in component_to_resolve.spark_input_column_names:
            return NLP_LEVELS.DOCUMENT

        # (2.) A model_anno_obj which is input dependent and not using document/sentence cols
        # We iterator over components and see which is feeding this input dependent component_to_resolve
        for c in pipe.components:
            if c.name == component_to_resolve.name:
                continue
            if c.spark_output_column_names[0] in component_to_resolve.spark_input_column_names:
                # We found a component that is feeding the component_to_resolve.
                # Now we need to check if that component is document/sentence level
                if NLP_LEVELS.DOCUMENT in c.spark_input_column_names:
                    return NLP_LEVELS.DOCUMENT
                elif NLP_LEVELS.SENTENCE in c.spark_input_column_names:
                    return NLP_LEVELS.SENTENCE

    @staticmethod
    def resolve_component_to_output_level(pipe, component) -> NlpLevel:
        """
        For a given NLU component_to_resolve, resolve its output level,
        by checking annotator_levels dicts for approaches and models
        If output level is input dependent, resolve_input_dependent_component_to_output_level will resolve it
        :param component:  to resolve
        :param pipe:  pipe containing the component
        :return: resolve component_to_resolve
        """
        if 'input_dependent' in component.output_level:
            return OutputLevelUtils.resolve_input_dependent_component_to_output_level(pipe, component)
        else:
            return component.output_level

    @staticmethod
    def get_columns_at_same_level_of_pipe(pipe, df, anno_2_ex_config, get_embeddings) -> List[str]:
        """Get List of columns in df that are generated from components in the pipeline
            which are at the same output level as the pipe .
        :param pipe: NLU Pipeline
        :param df: Pandas DataFrame resulting from applying the pipe
        :param anno_2_ex_config: mapping between anno to extractor, from get_annotator_extraction_configs()
        :param get_embeddings: Should embeddings be included
        :return: List of columns which are generated from components
                at same output level as the pipe.prediction_output_level
        """
        same_output_level_cols = []
        # if finisher exist, we must handle it first separately and blacklist all annos if it uses clean anno cols
        cleaned_components = []
        for c in pipe.components:
            if c.name == NLP_NODE_IDS.FINISHER:
                """
                Every col needs to have its own output level
                for every input col of finisher:
                1. Find provider component
                2. Resolve output level of provider component
                3. Use output level of provider component as output level of finisher col 
                """
                cleaned = c.model.getCleanAnnotations()
                get_meta = c.model.getIncludeMetadata()
                for col_in, col_out in zip(c.model.getInputCols(), c.model.getOutputCols()):
                    for provider_component in pipe.components:
                        if cleaned:
                            cleaned_components.append(provider_component)

                        m = provider_component.model
                        single_out = hasattr(m, 'getOutputCol')
                        if single_out and col_in in m.getOutputCol() or not single_out and col_in in m.getOutputCols():
                            output_level = OutputLevelUtils.resolve_component_to_output_level(pipe, provider_component)
                            if output_level == pipe.prediction_output_level:
                                # Normally need to use generated col, but finisher cols are not generated except meta?!
                                same_output_level_cols.append(col_out)
                                if get_meta:
                                    same_output_level_cols.append(col_out + '_metadata')
                continue

        for c in pipe.components:
            if c in cleaned_components:
                continue
            if 'embedding' in c.type and get_embeddings is False:
                continue
            if c.name == NLP_NODE_IDS.FINISHER:
                continue
            output_level = OutputLevelUtils.resolve_component_to_output_level(pipe, c)
            if output_level == pipe.prediction_output_level:
                generated_cols = ColSubstitutionUtils.get_final_output_cols_of_component(c, df, anno_2_ex_config)
                for generated_col in generated_cols:
                    if '_k_' in generated_col and c.jsl_anno_class_id == NLP_HC_NODE_IDS.SENTENCE_ENTITY_RESOLVER:
                        # all _k_ fields of resolver may never be viewed as any common outputlevel and thus never be zipped.
                        continue
                    same_output_level_cols.append(generated_col)
        return list(set(same_output_level_cols))

    @staticmethod
    def get_output_level_mapping_by_component(pipe) -> Dict[Any, str]:
        """Get a mapping key=NluComponent and value = output level
        :param pipe: NLU pipe for which to get the mapping
        :return: dict where key = NLU_Component and Value = Output level
        """
        nlp_levels = {c: OutputLevelUtils.resolve_component_to_output_level(pipe, c) for c in pipe.components}
        for c in pipe.components:
            if c.license == Licenses.ocr:
                nlp_levels[c] = c.output_level
        return {c: OutputLevelUtils.resolve_component_to_output_level(pipe, c) for c in pipe.components}
